import random
import re

import nltk
import spacy
from nltk import WordNetLemmatizer
from sentence_transformers import SentenceTransformer
from tqdm import tqdm

from .set_match_evaluation import eval_set_match
from .spice_evaluation import eval_spice
from .soft_spice_evaluation import *
from ..utils import is_graph_format, space_out_symbols_in_graph
import logging

# Set up logging configuration (adjust the level and format as needed)
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


class Evaluator:
    def __init__(self, parser=None, text_encoder_checkpoint=None, device='cuda:0', lemmatize=False):

        self.parser = parser
        if text_encoder_checkpoint:
            self.text_encoder = SentenceTransformer(text_encoder_checkpoint).to(device).eval()
        self.lemmatize = lemmatize

        if lemmatize:
            # # Load spacy model for lemmatization
            # self.nlp = spacy.load('en_core_web_sm', disable=['parser', 'ner'])
            #
            # # Define a custom token match function
            # def custom_token_match(text):
            #     # Match patterns like 'v:', 'pv:', ':1', ':2', etc.
            #     return re.match(r'(v:|pv:|:\d+)', text)
            #
            # # Update the tokenizer with the custom token match
            # self.nlp.tokenizer = Tokenizer(self.nlp.vocab, token_match=custom_token_match)
            self.lemmatizer = WordNetLemmatizer()

    def _process_graphs(self, graph_string):
        """
        Perform text processing: lemmatization.

        :param text: A string containing the text to be processed.
        :return: Processed text as a string.
        """



        if self.lemmatize:
            # Tokenize the text using spaCy
            # doc = self.nlp(graph_string)
            #
            # # Define a set of words to exclude from lemmatization
            # # Lemmatize tokens that are not in exclude_words and are not punctuations
            # lemmatized_tokens = [token.lemma_ if token.text not in exclude_words and not token.is_punct else token.text for token in doc]
            #
            # # Join the lemmatized tokens back into a string
            # graph_string = ' '.join(lemmatized_tokens)
            # Lemmatize each word in the text
            exclude_words = {'is', ',', '(', ')'}
            tokens = graph_string.split(' ')
            graph_string = ' '.join([self.lemmatizer.lemmatize(token) if not token in exclude_words else token for token in tokens])

        return graph_string

    def evaluate(self, candidates, references, method='spice', batch_size=4, return_graphs=False, **kwargs):
        """
        Evaluate scene graphs or text captions.

        :param candidates: List of candidate scene graphs or captions.
        :param references: List of List of reference scene graphs or captions.
        :param method: Evaluation method ('spice', 'soft_spice', or 'set_match').
        :param batch_size: Batch size for processing.
        :param kwargs: Additional arguments for evaluation metrics.
        :return: Evaluation scores, and optionally the processed candidates and references.
        """
        logging.info("Starting evaluation...")

        if not all(isinstance(candidate, str) for candidate in candidates):
            raise ValueError("All candidates must be strings.")

        # Ensure references is a list of lists of strings
        if not all(isinstance(ref_list, list) and all(isinstance(ref, str) for ref in ref_list) for ref_list in
                   references):
            raise ValueError("References must be a list of lists of strings.")

        # Determine input formats and parse if necessary
        candidates, references = self._parse_inputs(candidates, references, batch_size, **kwargs)

        # Choose the evaluation method
        method_function = {
            'set_match': self._set_match_score,
            'spice': self._spice_score,
            'soft_spice': self._soft_spice_score
        }.get(method)

        logging.info(f"Evaluating using method: {method}")

        if method_function is None:
            raise ValueError(f"Unknown evaluation method: {method}")

        # Filter kwargs based on the method
        if method == 'soft_spice':
            method_kwargs = {k: kwargs[k] for k in kwargs if k in ['bidirectional']}
            scores = method_function(candidates, references, batch_size, **method_kwargs)
        elif method == 'spice':
            method_kwargs = {k: kwargs[k] for k in kwargs if k in ['merge_tuples_synonyms', 'synonym_match']}
            scores = method_function(candidates, references, **method_kwargs)
        else:
            scores = method_function(candidates, references)

        logging.info("Evaluation completed.")

        # multiply with 100
        scores = [100 * score for score in scores]

        # Return results
        return (scores, candidates, references) if return_graphs else scores

    def _parse_inputs(self, candidates, references, batch_size, **kwargs):
        """
        Parse inputs if they are not in graph format.

        :param candidates: List of candidate scene graphs or captions.
        :param references: List of List of reference scene graphs or captions.
        :param batch_size: Batch size for processing.
        :param kwargs: Additional arguments for parsing.
        :return: Parsed candidates and references.
        """
        # Check for parser availability for non-graph formats
        if not self._all_items_are_graphs(candidates) and self.parser is None:
            raise ValueError("Parser is required for non-graph candidate inputs.")

        # Ensure the structure of references is correct
        assert all(isinstance(ref_list, list) for ref_list in
                   references), "Each reference should be a list of scene graphs or captions."

        # Parse candidates and references if they are not in graph format
        parsed_candidates = self._parse_if_needed(candidates, batch_size, is_nested=False, **kwargs)
        parsed_references = self._parse_if_needed(references, batch_size, is_nested=True, **kwargs)

        return parsed_candidates, parsed_references

    def _all_items_are_graphs(self, items):
        """
        Check if all items in a list are in graph format.

        :param items: List of items (candidates or references).
        :return: Boolean indicating if all items are in graph format.
        """
        return all(is_graph_format(item) for item in items)

    def _needs_parsing(self, items, is_nested, sample_size=5):
        """
        Determine if parsing is needed based on a random sample from a list of items.

        :param items: List of items or nested list of items.
        :param is_nested: Boolean indicating if 'items' is a nested list.
        :param sample_size: The number of items to sample for checking.
        :return: Boolean indicating if parsing is needed.
        """

        # Flatten the list if it is nested
        if is_nested:
            flattened_items = [item for sublist in items for item in sublist]
        else:
            flattened_items = items

        # Sample a few items from the list
        sampled_items = random.sample(flattened_items, min(sample_size, len(flattened_items)))

        return not any(is_graph_format(item) for item in sampled_items)

    def _parse_if_needed(self, items, batch_size, is_nested, **kwargs):
        """
        Parse items if they are not in graph format. Handles both nested and non-nested lists.
        Applies lemmatization to parsed graphs if enabled.

        :param items: List or list of lists of items (candidates or references).
        :param batch_size: Batch size for processing.
        :param is_nested: Boolean indicating if the items list is nested.
        :param kwargs: Additional arguments for parsing.
        :return: Parsed items, maintaining the original structure.
        """
        # Determine whether parsing is needed
        logging.info("Determine whether parsing is needed...")
        needs_parsing = self._needs_parsing(items, is_nested)
        logging.info(f"Parsing is needed: {needs_parsing}")
        if needs_parsing:
            if is_nested:
                logging.info("Parsing references...")
            else:
                logging.info("Parsing candidates...")

        # Flatten nested list if necessary and parse
        flat_list, structure = (self._flatten_nested_list(items) if is_nested else (items, None))
        parsed_flat_list = self.parser.parse(flat_list, batch_size=batch_size, return_text=True,
                                             **kwargs) if needs_parsing else flat_list

        # Apply lemmatization post-processing if enabled
        if self.lemmatize:
            parsed_flat_list = [self._process_graphs(graph_str) for graph_str in parsed_flat_list]

        parsed_flat_list = [space_out_symbols_in_graph(graph_str) for graph_str in parsed_flat_list]

        # Recover the nested list structure if it was flattened
        return self._recover_nested_list_structure(parsed_flat_list, structure) if is_nested else parsed_flat_list

    def _flatten_nested_list(self, nested_list):
        """
        Flatten a nested list while keeping track of the original structure.

        :param nested_list: A list of lists to be flattened.
        :return: A tuple of the flattened list and a list of lengths of the original sublists.
        """
        flat_list = []
        structure = []
        for sublist in nested_list:
            flat_list.extend(sublist)
            structure.append(len(sublist))
        return flat_list, structure

    def _recover_nested_list_structure(self, flat_list, structure):
        """
        Recover the structure of a nested list from the flattened version and the original structure information.

        :param flat_list: Flattened list of items.
        :param structure: List of lengths of the original sublists.
        :return: Nested list reconstructed from the flat list.
        """
        nested_list, index = [], 0
        for length in structure:
            nested_list.append(flat_list[index:index + length])
            index += length
        return nested_list

    def _set_match_score(self, candidates, references):
        """
        Set the match score for each candidate and reference pair.

        :param candidates: Candidate scene graphs.
        :param references: Reference scene graphs.
        :return: List of match scores.
        """
        scores = []
        for cand, refs in zip(candidates, references):
            score = eval_set_match(cand, refs)
            scores.append(score)
        return scores

    def _spice_score(self, candidates, references, merge_tuples_synonyms=False, synonym_match=True):
        """
        Compute SPICE score.

        :param candidates: List of Candidate scene graphs.
        :param references: List of List of Reference scene graphs.
        :return: List of SPICE scores.
        """
        scores = []
        for cand, refs in tqdm(zip(candidates, references), total=len(candidates)):
            score = eval_spice(cand, refs, merge_tuples_synonyms, synonym_match)
            scores.append(score)
        return scores

    def _soft_spice_score(self, candidates, references, batch_size, bidirectional=False):
        """
        Compute Soft SPICE scores for a batch of candidates and references.

        :param candidates: A list of candidate scene graphs.
        :param references: A list of reference scene graphs corresponding to the candidates.
        :param batch_size: Batch size to be used for encoding.
        :param bidirectional: Whether to use bidirectional encoding.
        :return: A list of Soft SPICE scores for each candidate.
        """
        print("using bidirectional: ", bidirectional)
        all_cand_phrases, all_ref_phrases, cand_lengths, ref_lengths = accumulate_phrases(candidates, references)
        encoded_cands, encoded_refs = encode_phrases(self.text_encoder, all_cand_phrases, all_ref_phrases, batch_size)
        scores = compute_scores(encoded_cands, encoded_refs, cand_lengths, ref_lengths, bidirectional)
        return scores






import logging
from importlib import resources

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

class SynonymDictionary:
    def __init__(self, exc_file_path, syn_file_path):
        self.word_to_bases = {}
        self.word_to_synsets = {}
        self.set_to_relations = {}

        # logging.info("Reading synonym dictionary..")
        # Reading exception file
        with open(exc_file_path, 'r', encoding='utf-8') as file:
            for base in file:
                base = base.strip()
                forms = next(file).strip().split()
                for form in forms:
                    self.word_to_bases.setdefault(form, []).append(base)

        # Reading synset file
        with open(syn_file_path, 'r', encoding='utf-8') as file:
            for word in file:
                word = word.strip()
                sets = set(map(int, next(file).strip().split()))
                self.word_to_synsets[word] = sets

        # logging.info("Reading synonym dictionary.. Done")
    def get_synsets(self, word):
        return self.word_to_synsets.get(word, set())

    def get_stem_synsets(self, word):
        bases = self.word_to_bases.get(word)
        if bases:
            sets = set()
            for base in bases:
                sets.update(self.get_synsets(base))
            return sets
        return self.get_synsets(self.morph(word))


    def morph(self, word):
        sufx = [
            "s", "ses", "xes", "zes", "ches", "shes", "men", "ies",  # Noun suffixes
            "s", "ies", "es", "es", "ed", "ed", "ing", "ing",        # Verb suffixes
            "er", "est", "er", "est"                                 # Adjective suffixes
        ]

        addr = [
            "", "s", "x", "z", "ch", "sh", "man", "y",               # Noun endings
            "", "y", "e", "", "e", "", "e", "",                      # Verb endings
            "", "", "e", "e"                                         # Adjective endings
        ]

        if word.endswith("ful"):
            base = word[:-3]  # Remove 'ful'
            if base in self.word_to_synsets:
                return base
            return word

        if word.endswith("ss") or len(word) <= 2:
            return word

        for sufx, addr in zip(sufx, addr):
            if word.endswith(sufx):
                base = word[:-len(sufx)] + addr
                if base in self.word_to_synsets:
                    return base

        return word


with resources.open_text('factual_scene_graph.evaluation.resources', 'english.exceptions') as f_exc, \
        resources.open_text('factual_scene_graph.evaluation.resources', 'english.synsets') as f_syn:
    synonym_dictionary = SynonymDictionary(f_exc.name,
                                            f_syn.name)





import os
import json
from typing import Dict, List, Any
import torch
from transformers import AutoTokenizer
from discosg_arch.T5ForDualTask import T5ForDualTask
from discosg.triple_utils import extract_triples

from tqdm import tqdm


class DualTaskSceneGraphParser:
    """
    DualTaskSceneGraphParser, functionalities:
    1. Insert triples into a scene graph based on a caption.
    2. Delete triples from a scene graph based on a caption.
    3. Delete triples from a scene graph based on a caption and then insert new triples.
    4. Insert triples into a scene graph based on a caption and then delete existing triples.
    """

    def __init__(
        self,
        model_path: str,
        device: str = "cuda",
        lemmatize: bool = False,
        lowercase: bool = True,
    ):
        """
        Initialize the DualTaskSceneGraphParser.

        Args:
            model_path: Path to the model directory.
            device: Device to run the model on (e.g., "cuda" or "cpu").
            lemmatize: Whether to perform lemmatization on the input text.
            lowercase: Whether to convert the input text to lowercase.
        """
        self.device = device
        self.lemmatize = lemmatize
        self.lowercase = lowercase

        # load model
        self.model = T5ForDualTask(model_name=model_path, use_lora=False, device=device)
        self.model.eval()

        # get base model name
        base_model_name = None
        model_info_path = os.path.join(model_path, "model_info.json")
        if os.path.exists(model_info_path):
            with open(model_info_path, "r") as f:
                model_info = json.load(f)
                base_model_name = model_info.get(
                    "base_model_name", "google/flan-t5-base"
                )
        else:
            base_model_name = "google/flan-t5-base"  # 默认模型

        # load tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(base_model_name)

        # load spacy if lemmatize is True
        if self.lemmatize:
            import spacy

            self.nlp = spacy.load("en_core_web_sm")

    def _preprocess_text(self, text: str) -> str:
        """preprocess the input text"""
        if self.lowercase:
            text = text.lower()
        if self.lemmatize:
            doc = self.nlp(text)
            text = " ".join([token.lemma_ for token in doc])
        return text

    def _extract_triples_from_text(self, text: str) -> List[List[str]]:
        """extract triples from the input text"""
        return extract_triples(text)

    def _delete_triples(self, scene_graph: str, caption: str, **kwargs) -> str:
        """execute delete task
        Args:
            scene_graph: The scene graph to delete triples from.
            caption: The caption to use for deletion.
        """
        triples = self._extract_triples_from_text(scene_graph)
        if not triples:
            return scene_graph

        input_text = f"Delete Task:\nCaption: {caption}\nCandidate Graph: {scene_graph}"

        encoding = self.tokenizer(
            input_text,
            max_length=512,
            padding="longest",
            truncation=True,
            return_tensors="pt",
            return_offsets_mapping=True,
        )

        offset_mapping = encoding.pop("offset_mapping")[0].tolist()
        input_ids = encoding["input_ids"].to(self.device)
        attention_mask = encoding["attention_mask"].to(self.device)

        triple_spans = []
        for triple_str in [f"( {' , '.join(triple)} )" for triple in triples]:
            triple_start_char = input_text.find(triple_str)
            if triple_start_char == -1:
                continue
            triple_end_char = triple_start_char + len(triple_str) - 1

            token_start = None
            token_end = None
            for i, (start, end) in enumerate(offset_mapping):
                if start == end == 0:
                    continue
                if token_start is None and start <= triple_start_char < end:
                    token_start = i
                if end > 0 and start <= triple_end_char < end:
                    token_end = i
                if (
                    start > triple_end_char
                    and token_end is None
                    and token_start is not None
                ):
                    token_end = i - 1
                    break

            if token_start is not None and token_end is not None:
                triple_spans.append([token_start, token_end])

        if not triple_spans:
            return scene_graph

        max_triples = 128
        spans_tensor = torch.full((1, max_triples, 2), -1, dtype=torch.long).to(
            self.device
        )
        spans_mask = torch.zeros(1, max_triples, dtype=torch.bool).to(self.device)

        for i, (start, end) in enumerate(triple_spans):
            if i >= max_triples:
                break
            spans_tensor[0, i, 0] = start
            spans_tensor[0, i, 1] = end
            spans_mask[0, i] = True

        with torch.no_grad():
            outputs = self.model.forward_delete(
                input_ids=input_ids,
                attention_mask=attention_mask,
                triple_spans=spans_tensor,
                span_mask=spans_mask,
            )

        logits = outputs["logits"]
        predictions = torch.argmax(logits, dim=1).cpu().numpy()

        kept_triples = []
        pred_idx = 0
        for i, _ in enumerate(triple_spans):
            if i >= len(predictions):
                kept_triples.append(triples[i])
                continue

            if predictions[pred_idx] == 0:  # KEEP
                kept_triples.append(triples[i])
            pred_idx += 1

        result = " , ".join([f"( {' , '.join(triple)} )" for triple in kept_triples])
        return result

    def _batch_delete_triples(
        self,
        scene_graphs: List[str],
        captions: List[str],
        max_input_len=2048,
        max_triples_num=128,
        skip_toolong=False,
        skip_len=-1,
        **kwargs,
    ) -> List[str]:
        if not scene_graphs:
            return []

        # assert when skip_toolong is True, skip_len should be > 0
        # if skip_toolong and skip_len <= 0:
        #     raise ValueError("When skip_toolong is True, skip_len should be > 0")

        batch_size = len(scene_graphs)
        results = [""] * batch_size

        all_triples = []
        valid_indices = []

        for i, scene_graph in enumerate(scene_graphs):
            triples = self._extract_triples_from_text(scene_graph)
            if triples:
                all_triples.append(triples)
                valid_indices.append(i)
            else:
                results[i] = scene_graph

        if not valid_indices:
            return results

        valid_scene_graphs = [scene_graphs[i] for i in valid_indices]
        valid_captions = [captions[i] for i in valid_indices]
        input_texts = [
            f"Delete Task:\nCaption: {caption}\nCandidate Graph: {graph}"
            for caption, graph in zip(valid_captions, valid_scene_graphs)
        ]

        # encode the input texts
        encodings = self.tokenizer(
            input_texts,
            max_length=max_input_len,
            padding="longest",
            truncation=True,
            return_tensors="pt",
            return_offsets_mapping=True,
        )

        offset_mappings = encodings.pop("offset_mapping").tolist()
        input_ids = encodings["input_ids"].to(self.device)
        attention_mask = encodings["attention_mask"].to(self.device)

        # set max_triples: we process 128 triples at most at a time
        max_triples = max_triples_num
        # get triple spans
        batch_triple_spans = []

        for batch_idx, (input_text, triples, offset_mapping) in enumerate(
            zip(input_texts, all_triples, offset_mappings)
        ):
            triple_spans = []
            for triple_str in [f"( {' , '.join(triple)} )" for triple in triples]:
                triple_start_char = input_text.find(triple_str)
                if triple_start_char == -1:
                    continue
                triple_end_char = triple_start_char + len(triple_str) - 1

                token_start = None
                token_end = None
                for i, (start, end) in enumerate(offset_mapping):
                    if start == end == 0:
                        continue
                    if token_start is None and start <= triple_start_char < end:
                        token_start = i
                    if end > 0 and start <= triple_end_char < end:
                        token_end = i
                    if (
                        start > triple_end_char
                        and token_end is None
                        and token_start is not None
                    ):
                        token_end = i - 1
                        break

                if token_start is not None and token_end is not None:
                    triple_spans.append([token_start, token_end])

            batch_triple_spans.append(triple_spans)

        # prepare spans tensor and mask for model input
        spans_tensor = torch.full(
            (len(valid_indices), max_triples, 2), -1, dtype=torch.long
        ).to(self.device)
        spans_mask = torch.zeros(len(valid_indices), max_triples, dtype=torch.bool).to(
            self.device
        )

        for batch_idx, triple_spans in enumerate(batch_triple_spans):
            for i, (start, end) in enumerate(triple_spans):
                if i >= max_triples:
                    break
                spans_tensor[batch_idx, i, 0] = start
                spans_tensor[batch_idx, i, 1] = end
                spans_mask[batch_idx, i] = True

        # forward pass
        with torch.no_grad():
            outputs = self.model.forward_delete(
                input_ids=input_ids,
                attention_mask=attention_mask,
                triple_spans=spans_tensor,
                span_mask=spans_mask,
            )

        # get logits
        logits = outputs["logits"]

        triple_counts = [min(len(spans), max_triples) for spans in batch_triple_spans]

        all_predictions = torch.argmax(logits, dim=1).cpu().numpy()

        start_idx = 0
        for batch_idx, count in enumerate(triple_counts):
            if count == 0:
                continue

            sample_predictions = all_predictions[start_idx : start_idx + count]
            start_idx += count

            kept_triples = []
            for i, pred in enumerate(sample_predictions):
                if i < len(all_triples[batch_idx]) and pred == 0:  # KEEP
                    kept_triples.append(all_triples[batch_idx][i])

            result = " , ".join(
                [f"( {' , '.join(triple)} )" for triple in kept_triples]
            )
            results[valid_indices[batch_idx]] = result

        # debug
        for result in results:
            if result == "" or len(result) < 5:
                print("Empty result")

        # skip too long encoded encodings
        # TODO: filter out too long encodings before encoding and ner deletion
        # if skip_toolong:
        #     # calculate the valid length of encoding, using sum(mask)
        #     # for i, scene_graph, encoding in enumerate(zip(scene_graphs, encodings["input_ids"])):
        #     for i, scene_graph, encoding, mask in zip(
        #         valid_indices,
        #         scene_graphs,
        #         encodings["input_ids"],
        #         encodings["attention_mask"],
        #     ):
        #         if sum(mask) > skip_len:
        #             results[i] = scene_graph

        return results

    def _batch_delete_triples_skip_len(
        self,
        scene_graphs: List[str],
        captions: List[str],
        max_input_len=2048,
        max_triples_num=128,
        skip_toolong=False,
        skip_len=-1,
        **kwargs,
    ) -> List[str]:
        if not scene_graphs:
            return []

        # assert when skip_toolong is True, skip_len should be > 0
        if skip_toolong and skip_len <= 0:
            raise ValueError("When skip_toolong is True, skip_len should be > 0")

        batch_size = len(scene_graphs)
        results = [""] * batch_size

        all_triples = []
        valid_indices = []

        for i, scene_graph in enumerate(scene_graphs):
            triples = self._extract_triples_from_text(scene_graph)
            if triples:
                all_triples.append(triples)
                valid_indices.append(i)
            else:
                results[i] = scene_graph

        if not valid_indices:
            return results

        valid_scene_graphs = [scene_graphs[i] for i in valid_indices]
        valid_captions = [captions[i] for i in valid_indices]
        input_texts = [
            f"Delete Task:\nCaption: {caption}\nCandidate Graph: {graph}"
            for caption, graph in zip(valid_captions, valid_scene_graphs)
        ]

        # encode the input texts
        encoding_for_selection = self.tokenizer(
            input_texts,
            max_length=max_input_len * 4,
            padding="longest",
            truncation=True,
            return_tensors="pt",
        ).to(self.device)
        encodings = self.tokenizer(
            input_texts,
            max_length=max_input_len * 4,
            padding="longest",
            truncation=True,
            return_tensors="pt",
            return_offsets_mapping=True,
        )

        offset_mappings = encodings.pop("offset_mapping").tolist()
        input_ids = encodings["input_ids"].to(self.device)
        attention_mask = encodings["attention_mask"].to(self.device)
        attention_mask_for_selection = encoding_for_selection["attention_mask"].to(
            self.device
        )

        if skip_toolong:
            skip_indices = []
            for batch_idx, mask in enumerate(attention_mask_for_selection):
                # print(sum(mask).item())
                if sum(mask).item() > skip_len:
                    original_idx = valid_indices[batch_idx]
                    results[original_idx] = scene_graphs[original_idx]
                    skip_indices.append(batch_idx)

            if skip_indices:
                keep_indices = [
                    i for i in range(len(valid_indices)) if i not in skip_indices
                ]
                if not keep_indices:
                    return results

                input_ids = input_ids[keep_indices]
                attention_mask = attention_mask[keep_indices]
                offset_mappings = [offset_mappings[i] for i in keep_indices]
                input_texts = [input_texts[i] for i in keep_indices]
                all_triples = [all_triples[i] for i in keep_indices]
                valid_indices = [valid_indices[i] for i in keep_indices]

        # set max_triples: we process max_triples_num triples at most at a time
        max_triples = max_triples_num
        # get triple spans
        batch_triple_spans = []

        for batch_idx, (input_text, triples, offset_mapping) in enumerate(
            zip(input_texts, all_triples, offset_mappings)
        ):
            triple_spans = []
            for triple_str in [f"( {' , '.join(triple)} )" for triple in triples]:
                triple_start_char = input_text.find(triple_str)
                if triple_start_char == -1:
                    continue
                triple_end_char = triple_start_char + len(triple_str) - 1

                token_start = None
                token_end = None
                for i, (start, end) in enumerate(offset_mapping):
                    if start == end == 0:
                        continue
                    if token_start is None and start <= triple_start_char < end:
                        token_start = i
                    if end > 0 and start <= triple_end_char < end:
                        token_end = i
                    if (
                        start > triple_end_char
                        and token_end is None
                        and token_start is not None
                    ):
                        token_end = i - 1
                        break

                if token_start is not None and token_end is not None:
                    triple_spans.append([token_start, token_end])

            batch_triple_spans.append(triple_spans)

        # prepare spans tensor and mask for model input
        spans_tensor = torch.full(
            (len(valid_indices), max_triples, 2), -1, dtype=torch.long
        ).to(self.device)
        spans_mask = torch.zeros(len(valid_indices), max_triples, dtype=torch.bool).to(
            self.device
        )

        for batch_idx, triple_spans in enumerate(batch_triple_spans):
            for i, (start, end) in enumerate(triple_spans):
                if i >= max_triples:
                    break
                spans_tensor[batch_idx, i, 0] = start
                spans_tensor[batch_idx, i, 1] = end
                spans_mask[batch_idx, i] = True

        # forward pass
        with torch.no_grad():
            outputs = self.model.forward_delete(
                input_ids=input_ids,
                attention_mask=attention_mask,
                triple_spans=spans_tensor,
                span_mask=spans_mask,
            )

        # get logits
        logits = outputs["logits"]

        triple_counts = [min(len(spans), max_triples) for spans in batch_triple_spans]

        # pred label for each triple, 0 for KEEP, 1 for DELETE
        all_predictions = torch.argmax(logits, dim=1).cpu().numpy()

        # allocate the predictions to each sample
        start_idx = 0
        for batch_idx, count in enumerate(triple_counts):
            if count == 0:
                continue

            sample_predictions = all_predictions[start_idx : start_idx + count]
            start_idx += count

            kept_triples = []
            for i, pred in enumerate(sample_predictions):
                if i < len(all_triples[batch_idx]) and pred == 0:  # KEEP
                    kept_triples.append(all_triples[batch_idx][i])

            result = " , ".join(
                [f"( {' , '.join(triple)} )" for triple in kept_triples]
            )
            results[valid_indices[batch_idx]] = result

        # debug
        # for i, result in enumerate(results):
        #     if result == "" or len(result) < 5:
        #         print("\nEmpty result, replace with original scene graph\n")
        #         results[i] = scene_graphs[valid_indices[i]] # TODO：check

        return results

    def _batch_delete_triples_maskBIO(
        self, scene_graphs: List[str], captions: List[str], **kwargs
    ) -> List[str]:
        if not scene_graphs:
            return []

        batch_size = len(scene_graphs)
        results = [""] * batch_size

        all_triples = []
        valid_indices = []

        for i, scene_graph in enumerate(scene_graphs):
            triples = self._extract_triples_from_text(scene_graph)
            if triples:
                all_triples.append(triples)
                valid_indices.append(i)
            else:
                results[i] = scene_graph

        if not valid_indices:
            return results

        valid_scene_graphs = [scene_graphs[i] for i in valid_indices]
        valid_captions = [captions[i] for i in valid_indices]
        input_texts = [
            f"Delete Task:\nCaption: {caption}\nCandidate Graph: {graph}"
            for caption, graph in zip(valid_captions, valid_scene_graphs)
        ]

        encodings = self.tokenizer(
            input_texts,
            max_length=512,
            padding="longest",
            truncation=True,
            return_tensors="pt",
            return_offsets_mapping=True,
        )

        input_ids = encodings["input_ids"].to(self.device)
        attention_mask = encodings["attention_mask"].to(self.device)
        offset_mapping = encodings["offset_mapping"].to(self.device)

        max_triples = 128
        valid_batch_size = len(valid_indices)

        spans_tensor = torch.full(
            (valid_batch_size, max_triples, 2), -1, dtype=torch.long, device=self.device
        )
        spans_mask = torch.zeros(
            (valid_batch_size, max_triples), dtype=torch.bool, device=self.device
        )

        for batch_idx, (input_text, triples) in enumerate(
            zip(input_texts, all_triples)
        ):
            if not triples:
                continue

            triple_strings = [f"( {' , '.join(triple)} )" for triple in triples]

            for i, triple_str in enumerate(triple_strings):
                if i >= max_triples:
                    break

                triple_start_char = input_text.find(triple_str)
                if triple_start_char == -1:
                    continue
                triple_end_char = triple_start_char + len(triple_str) - 1

                sample_offset = offset_mapping[batch_idx]

                char_starts = sample_offset[:, 0]
                char_ends = sample_offset[:, 1]

                start_mask = (
                    (char_starts <= triple_start_char)
                    & (triple_start_char < char_ends)
                    & (char_ends > 0)
                )
                if not torch.any(start_mask):
                    continue
                token_start = torch.nonzero(start_mask, as_tuple=True)[0][0].item()

                end_mask = (
                    (char_starts <= triple_end_char)
                    & (triple_end_char < char_ends)
                    & (char_ends > 0)
                )
                if not torch.any(end_mask):
                    end_mask = (
                        (char_ends > 0)
                        & (char_starts > triple_start_char)
                        & (char_starts <= triple_end_char)
                    )
                    if not torch.any(end_mask):
                        continue
                    token_end = torch.nonzero(end_mask, as_tuple=True)[0][-1].item()
                else:
                    token_end = torch.nonzero(end_mask, as_tuple=True)[0][0].item()

                spans_tensor[batch_idx, i, 0] = token_start
                spans_tensor[batch_idx, i, 1] = token_end
                spans_mask[batch_idx, i] = True

        if not torch.any(spans_mask):
            return results

        with torch.no_grad():
            outputs = self.model.forward_delete_maskBIO(
                input_ids=input_ids,
                attention_mask=attention_mask,
                triple_spans=spans_tensor,
                span_mask=spans_mask,
            )

        logits = outputs["logits"]
        predictions = torch.argmax(logits, dim=1).cpu()

        batch_indices = outputs.get("batch_indices", None)

        if batch_indices is not None:
            batch_indices = batch_indices.cpu()

            batch_predictions = [[] for _ in range(valid_batch_size)]

            for i, batch_idx in enumerate(batch_indices):
                batch_idx = batch_idx.item()
                if batch_idx < valid_batch_size:
                    batch_predictions[batch_idx].append(predictions[i].item())

            for batch_idx, (preds, triples) in enumerate(
                zip(batch_predictions, all_triples)
            ):
                kept_triples = [
                    triples[i]
                    for i, pred in enumerate(preds)
                    if i < len(triples) and pred == 0
                ]
                result = " , ".join(
                    [f"( {' , '.join(triple)} )" for triple in kept_triples]
                )
                results[valid_indices[batch_idx]] = result
        else:
            triple_counts = spans_mask.sum(dim=1).cpu().tolist()

            start_idx = 0
            for batch_idx, count in enumerate(triple_counts):
                count = int(count)
                if count == 0:
                    continue

                sample_predictions = predictions[start_idx : start_idx + count].tolist()
                start_idx += count

                kept_triples = []
                for i, pred in enumerate(sample_predictions):
                    if i < len(all_triples[batch_idx]) and pred == 0:  # KEEP
                        kept_triples.append(all_triples[batch_idx][i])

                result = " , ".join(
                    [f"( {' , '.join(triple)} )" for triple in kept_triples]
                )
                results[valid_indices[batch_idx]] = result

        return results

    def _insert_triples(self, scene_graph: str, caption: str, **kwargs) -> str:
        """执行插入任务"""
        input_text = f"Insert Task:\nCaption: {caption}\nCorrupted Graph: {scene_graph}"

        input_encoding = self.tokenizer(
            input_text,
            max_length=512,
            padding="longest",
            truncation=True,
            return_tensors="pt",
        ).to(self.device)

        with torch.no_grad():
            generated_ids = self.model.generate(
                input_ids=input_encoding["input_ids"],
                attention_mask=input_encoding["attention_mask"],
                max_length=512,
                num_beams=3,
                num_return_sequences=1,
                temperature=0.7,
                top_k=50,
                top_p=0.95,
                do_sample=True,
            )

        generated_text = self.tokenizer.decode(
            generated_ids[0], skip_special_tokens=True
        )

        if generated_text.strip():
            original_triples = self._extract_triples_from_text(scene_graph)
            generated_triples = self._extract_triples_from_text(generated_text)

            all_triples = original_triples.copy()
            for triple in generated_triples:
                if triple not in all_triples:
                    all_triples.append(triple)

            result = " , ".join([f"( {' , '.join(triple)} )" for triple in all_triples])
            return result

        return scene_graph

    def _batch_insert_triples(
        self,
        scene_graphs: List[str],
        captions: List[str],
        max_input_len=2048,
        max_output_len=512,
        skip_toolong=False,
        skip_len=-1,
        **kwargs,
    ) -> List[str]:
        if not scene_graphs:
            return []

        batch_size = len(scene_graphs)
        results = [""] * batch_size

        valid_indices = []
        valid_scene_graphs = []
        valid_captions = []

        for i, (scene_graph, caption) in enumerate(zip(scene_graphs, captions)):
            if scene_graph.strip():
                valid_indices.append(i)
                valid_scene_graphs.append(scene_graph)
                valid_captions.append(caption)
            else:
                results[i] = ""

        if not valid_indices:
            return results

        input_texts = [
            f"Insert Task:\nCaption: {caption}\nCorrupted Graph: {scene_graph}"
            for caption, scene_graph in zip(valid_captions, valid_scene_graphs)
        ]

        input_encoding = self.tokenizer(
            input_texts,
            max_length=max_input_len,
            padding="longest",
            truncation=True,
            return_tensors="pt",
        ).to(self.device)

        with torch.no_grad():
            generated_ids = self.model.generate(
                input_ids=input_encoding["input_ids"],
                attention_mask=input_encoding["attention_mask"],
                max_length=max_output_len,
                num_beams=3,
                num_return_sequences=1,
                temperature=0.7,
                top_k=50,
                top_p=0.95,
                do_sample=True,
            )

        generated_texts = [
            self.tokenizer.decode(ids, skip_special_tokens=True)
            for ids in generated_ids
        ]

        for i, (scene_graph, generated_text, idx) in enumerate(
            zip(valid_scene_graphs, generated_texts, valid_indices)
        ):
            if generated_text.strip():
                original_triples = self._extract_triples_from_text(scene_graph)
                generated_triples = self._extract_triples_from_text(generated_text)

                all_triples = original_triples.copy()
                for triple in generated_triples:
                    if triple not in all_triples:
                        all_triples.append(triple)

                result = " , ".join(
                    [f"( {' , '.join(triple)} )" for triple in all_triples]
                )
                results[idx] = result
            else:
                results[idx] = scene_graph

        # skip too long encoded encodings
        # TODO: filter out too long encodings before encoding and ner deletion
        if skip_toolong:
            # calculate the valid length of encoding, using sum(mask)
            # for i, scene_graph, encoding in enumerate(zip(scene_graphs, encodings["input_ids"])):
            for i, scene_graph, encoding, mask in zip(
                valid_indices,
                scene_graphs,
                input_encoding["input_ids"],
                input_encoding["attention_mask"],
            ):
                if sum(mask) > skip_len:
                    results[i] = scene_graph

        return results

    def _batch_insert_triples_skip_len(
        self,
        scene_graphs: List[str],
        captions: List[str],
        max_input_len=2048,
        max_output_len=512,
        skip_toolong=False,
        skip_len=-1,
        num_beams=1,
        do_sample=True,
        top_k=50,
        top_p=1.0,
        temperature=1.0,
        **kwargs,
    ) -> List[str]:
        """Batch execution of insert tasks"""
        if not scene_graphs:
            return []

        # Ensure skip_len is valid when skip_toolong is True
        if skip_toolong and skip_len <= 0:
            raise ValueError("When skip_toolong is True, skip_len should be > 0")

        batch_size = len(scene_graphs)
        results = [""] * batch_size

        # Filter out empty scene graphs
        valid_indices = []
        valid_scene_graphs = []
        valid_captions = []

        for i, (scene_graph, caption) in enumerate(zip(scene_graphs, captions)):
            if scene_graph.strip():
                valid_indices.append(i)
                valid_scene_graphs.append(scene_graph)
                valid_captions.append(caption)
            else:
                results[i] = ""  # Return empty string for empty scene graphs

        if not valid_indices:
            return results

        # Prepare batch inputs
        input_texts = [
            f"Insert Task:\nCaption: {caption}\nCorrupted Graph: {scene_graph}"
            for caption, scene_graph in zip(valid_captions, valid_scene_graphs)
        ]

        # Encode input texts with max_length + 1 for select the over length samples
        input_encoding_for_selection = self.tokenizer(
            input_texts,
            max_length=max_input_len * 4,
            padding="longest",
            truncation=True,
            return_tensors="pt",
        ).to(self.device)
        # Encode input texts
        input_encoding = self.tokenizer(
            input_texts,
            max_length=max_input_len,
            padding="longest",
            truncation=True,
            return_tensors="pt",
        ).to(self.device)

        # Check encoding length and skip too long samples before model processing
        if skip_toolong:
            skip_indices = []
            for batch_idx, mask in enumerate(
                input_encoding_for_selection["attention_mask"]
            ):
                if sum(mask).item() > skip_len:
                    original_idx = valid_indices[batch_idx]
                    results[original_idx] = scene_graphs[original_idx]
                    skip_indices.append(batch_idx)
                    print(
                        f"Skipping sample {original_idx} due to length: {sum(mask).item()}"
                    )

            # Remove skipped samples from processing data
            if skip_indices:
                # Create indices for samples to keep
                keep_indices = [
                    i for i in range(len(valid_indices)) if i not in skip_indices
                ]
                if not keep_indices:  # If all samples are skipped
                    return results

                # Update all relevant data structures
                input_ids = input_encoding["input_ids"][keep_indices]
                attention_mask = input_encoding["attention_mask"][keep_indices]
                input_encoding = {
                    "input_ids": input_ids,
                    "attention_mask": attention_mask,
                }

                # Update valid_indices and other related lists
                valid_scene_graphs = [valid_scene_graphs[i] for i in keep_indices]
                valid_indices = [valid_indices[i] for i in keep_indices]

        # Return early if no valid samples remain
        if len(valid_indices) == 0:
            return results

        # Generate triples in batch
        with torch.no_grad():
            generated_ids = self.model.generate(
                input_ids=input_encoding["input_ids"],
                attention_mask=input_encoding["attention_mask"],
                max_length=max_output_len,
                num_beams=num_beams,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                do_sample=do_sample,
                num_return_sequences=1,
            )
            # generated_ids = self.model.generate(
            #     input_ids=input_encoding["input_ids"],
            #     attention_mask=input_encoding["attention_mask"],
            #     max_length=max_output_len,
            #     num_beams=3,
            #     temperature=0.7,
            #     top_k=50,
            #     top_p=0.95,
            #     do_sample=True,
            #     num_return_sequences=1,
            # )

        # Decode generated text
        generated_texts = [
            self.tokenizer.decode(ids, skip_special_tokens=True)
            for ids in generated_ids
        ]

        # Process each scene graph
        for i, (scene_graph, generated_text, idx) in enumerate(
            zip(valid_scene_graphs, generated_texts, valid_indices)
        ):
            # Add generated triples to original scene graph if not empty
            if generated_text.strip() and len(generated_text.strip()) > 5:
                # Extract original and generated triples
                original_triples = self._extract_triples_from_text(scene_graph)
                generated_triples = self._extract_triples_from_text(generated_text)

                # Merge triples (remove duplicates)
                all_triples = original_triples.copy()
                for triple in generated_triples:
                    if triple not in all_triples:
                        all_triples.append(triple)

                # Convert merged triples back to string format
                result = " , ".join(
                    [f"( {' , '.join(triple)} )" for triple in all_triples]
                )
                results[idx] = result
            else:
                print(
                    f"Generated text is empty or too short: {generated_text.strip()}, replacing with original scene graph"
                )
                results[idx] = scene_graph

        return results

    def _insert_iter(self, desc, graph_to_fix, batch_size):
        results = {}
        for i in range(0, len(desc), batch_size):
            batch_desc = desc[i : i + batch_size]
            for _, d in enumerate(batch_desc):
                if graph_to_fix and d in graph_to_fix:
                    scene_graph = graph_to_fix[d]
                    results[d] = self._insert_triples(
                        scene_graph=scene_graph, caption=d
                    )
                else:
                    results[d] = ""
        return results

    def _delete_iter(self, desc, graph_to_fix, batch_size):
        results = {}
        for i in range(0, len(desc), batch_size):
            batch_desc = desc[i : i + batch_size]
            for _, d in enumerate(batch_desc):
                if graph_to_fix and d in graph_to_fix:
                    scene_graph = graph_to_fix[d]
                    results[d] = self._batch_delete_triples(
                        scene_graph=scene_graph, caption=d
                    )
                else:
                    results[d] = ""
        return results

    def parse(
        self,
        descriptions: List[str],
        graph_to_fix=None,
        batch_size: int = 32,
        task: str = "insert_delete",
        max_input_len: int = 2048,
        max_output_len: int = 512,
        max_triples_num: int = 128,
        skip_toolong=False,
        skip_len=-1,
        num_beams: int = 1,
        do_sample: bool = False,
        top_k: int = 50,
        top_p: float = 1.0,
        temperature: float = 1.0,
    ) -> Dict[str, Any]:

        results = {}

        if task == "delete":
            results = {}
            for i in range(0, len(descriptions), batch_size):
                batch_descriptions = descriptions[
                    i : i + batch_size
                ]  # TODO: need batchify
                for j, desc in enumerate(batch_descriptions):
                    if graph_to_fix and desc in graph_to_fix:
                        scene_graph = graph_to_fix[desc]
                        results[desc] = self._batch_delete_triples(
                            scene_graph=scene_graph,
                            caption=desc,
                            max_input_len=max_input_len * 4,
                            max_triples_num=max_triples_num,
                        )
                    else:
                        results[desc] = ""
            return results

        elif task == "insert":  # TODO: check here
            results = {}
            for i in range(0, len(descriptions), batch_size):
                batch_descriptions = descriptions[i : i + batch_size]
                for j, desc in enumerate(batch_descriptions):
                    if graph_to_fix and desc in graph_to_fix:
                        scene_graph = graph_to_fix[desc]
                        results[desc] = self._batch_insert_triples_skip_len(
                            scene_graph=scene_graph,
                            caption=desc,
                            skip_toolong=skip_toolong,
                            skip_len=skip_len,
                            max_input_len=max_input_len,
                            max_output_len=max_output_len,
                            num_beams=num_beams if not do_sample else 1,
                            do_sample=do_sample,
                            top_k=top_k if do_sample else 50,
                            top_p=top_p if do_sample else 1.0,
                            temperature=temperature if do_sample else 1.0,
                        )
                    else:
                        results[desc] = ""
            return results

        elif task == "insert_delete":
            delete_results = {}
            for i in tqdm(range(0, len(descriptions), batch_size), ncols=88):
                batch_descriptions = descriptions[i : i + batch_size]
                batch_scene_graphs = []
                batch_captions = []

                for desc in batch_descriptions:
                    if graph_to_fix and desc in graph_to_fix:
                        batch_scene_graphs.append(graph_to_fix[desc])
                        batch_captions.append(desc)
                    else:
                        batch_scene_graphs.append("")
                        batch_captions.append(desc)  # TODO: check here

                batch_delete_results = self._batch_delete_triples(
                    scene_graphs=batch_scene_graphs,
                    captions=batch_captions,
                    max_input_len=max_input_len,
                    max_triples_num=max_triples_num,
                )
                for batch_idx, desc in enumerate(batch_captions):
                    delete_results[desc] = batch_delete_results[batch_idx]

            insert_results = {}
            for i in tqdm(range(0, len(descriptions), batch_size), ncols=88):
                batch_descriptions = descriptions[i : i + batch_size]
                batch_scene_graphs = []
                batch_captions = []
                for desc in batch_descriptions:
                    if graph_to_fix and desc in graph_to_fix:
                        batch_scene_graphs.append(graph_to_fix[desc])
                        batch_captions.append(desc)
                    else:
                        batch_scene_graphs.append("")
                        batch_captions.append(desc)
                batch_insert_results = self._batch_insert_triples_skip_len(
                    scene_graphs=batch_scene_graphs,
                    captions=batch_captions,
                    skip_toolong=skip_toolong,
                    skip_len=skip_len,
                    max_input_len=max_input_len,
                    max_output_len=max_output_len,
                    num_beams=num_beams if not do_sample else 1,
                    do_sample=do_sample,
                    top_k=top_k if do_sample else 50,
                    top_p=top_p if do_sample else 1.0,
                    temperature=temperature if do_sample else 1.0,
                )
                for batch_idx, desc in enumerate(batch_captions):
                    insert_results[desc] = batch_insert_results[batch_idx]

            return {"delete": delete_results, "insert": insert_results}

        elif task == "delete_before_insert":
            # insert based on delete results
            delete_results = {}
            for i in tqdm(range(0, len(descriptions), batch_size), ncols=88):
                batch_descriptions = descriptions[i : i + batch_size]
                batch_scene_graphs = []
                batch_captions = []
                for desc in batch_descriptions:
                    if graph_to_fix and desc in graph_to_fix:
                        batch_scene_graphs.append(graph_to_fix[desc])
                        batch_captions.append(desc)
                    else:
                        batch_scene_graphs.append("")
                        batch_captions.append(desc)
                batch_delete_results = self._batch_delete_triples(
                    scene_graphs=batch_scene_graphs,
                    captions=batch_captions,
                    max_input_len=max_input_len,
                    max_triples_num=max_triples_num,
                )
                for batch_idx, desc in enumerate(batch_captions):
                    delete_results[desc] = batch_delete_results[batch_idx]

            insert_results = {}

            for i in tqdm(range(0, len(descriptions), batch_size), ncols=88):
                batch_descriptions = descriptions[i : i + batch_size]
                batch_scene_graphs = []
                batch_captions = []
                for desc in batch_descriptions:
                    if delete_results and desc in delete_results:
                        batch_scene_graphs.append(delete_results[desc])
                        batch_captions.append(desc)
                    else:
                        batch_scene_graphs.append("")
                        batch_captions.append(desc)
                batch_insert_results = self._batch_insert_triples_skip_len(
                    scene_graphs=batch_scene_graphs,
                    captions=batch_captions,
                    skip_toolong=skip_toolong,
                    skip_len=skip_len,
                    max_input_len=max_input_len,
                    max_output_len=max_output_len,
                    num_beams=num_beams if not do_sample else 1,
                    do_sample=do_sample,
                    top_k=top_k if do_sample else 50,
                    top_p=top_p if do_sample else 1.0,
                    temperature=temperature if do_sample else 1.0,
                )
                for batch_idx, desc in enumerate(batch_captions):
                    insert_results[desc] = batch_insert_results[batch_idx]
            return {"delete": delete_results, "insert": insert_results}

        else:
            raise ValueError(f"Unsupported task: {task}")


# def main():
#     # Example usage
#     model = DualTaskSceneGraphParser(model_path="sqlinn/DiscoSG-Refiner-Large-t5-only", device="cuda", lemmatize=False, lowercase=True)
#     descpriptions = [
#         "The image captures a bustling urban scene, likely in a European city. The setting appears to be a pedestrian-friendly square or plaza. There are numerous people of various ages and attire walking around, some carrying bags, suggesting shopping or a day out. A few individuals are seated, possibly enjoying a meal or resting. The square is adorned with a decorative fountain in the center, surrounded by potted plants. Overhead, there are power lines and cables, hinting at an urban environment. The architecture of the surrounding buildings suggests a historic or older part of the city.",
#         "In the image, a man is seated at a desk, engrossed in his work on a computer. He's wearing a blue shirt and glasses, and his hand is raised to his forehead in a gesture that suggests deep thought or concentration. The desk, cluttered with various items, houses a computer monitor, keyboard, and mouse. The room around him is dimly lit, creating an atmosphere of focus and seriousness. In the background, a window can be seen, adding depth to the scene. The image captures a moment of intense concentration and productivity.",
#     ]
#     graph_to_fix = {
#         "The image captures a bustling urban scene, likely in a European city. The setting appears to be a pedestrian-friendly square or plaza. There are numerous people of various ages and attire walking around, some carrying bags, suggesting shopping or a day out. A few individuals are seated, possibly enjoying a meal or resting. The square is adorned with a decorative fountain in the center, surrounded by potted plants. Overhead, there are power lines and cables, hinting at an urban environment. The architecture of the surrounding buildings suggests a historic or older part of the city.": "( city , is , bustling ) , ( city , is , european ) , ( setting , is , pedestrian-friendly ) , ( setting , is , square ) , ( people , carry , bags ) , ( people , is , walking ) , ( individuals , is , seated ) , ( fountain , in center of , square ) , ( fountain , is , decorative ) , ( plants , is , potted ) , ( plants , surround , fountain ) , ( cables , is , overhead ) , ( power lines , is , overhead ) , ( buildings , surround , city ) , ( city , is , historic ) , ( city , is , older )",
#         "In the image, a man is seated at a desk, engrossed in his work on a computer. He's wearing a blue shirt and glasses, and his hand is raised to his forehead in a gesture that suggests deep thought or concentration. The desk, cluttered with various items, houses a computer monitor, keyboard, and mouse. The room around him is dimly lit, creating an atmosphere of focus and seriousness. In the background, a window can be seen, adding depth to the scene. The image captures a moment of intense concentration and productivity.": "( man , sit at , desk ) , ( man , work on , computer ) , ( hand , lift to , forehead ) , ( man , have , hand ) , ( man , wear , glasses ) , ( shirt , is , blue ) , ( desk , house , monitor ) , ( desk , house , mouse ) , ( desk , is , cluttered ) , ( monitor , is , computer ) , ( man , in , room ) , ( room , is , dimly lit ) , ( window , in , background ) , ( image , capture , concentration ) , ( image , capture , productivity ) , ( productivity , is , intense )",
#     }

#     outputs = model.parse(
#         descriptions=descpriptions,
#         graph_to_fix=graph_to_fix,
#         batch_size=2,
#         task="delete_before_insert",
#     )
#     print(outputs)
#     print()
#     print(outputs.keys())

# if __name__ == "__main__":
#     main()


import numpy as np
from packaging import version
from ..utils import get_seg_list

def _get_graph_phrases(graph_str_list, type_dict):
    """
    Extract phrases from graph strings and classify their types.

    :param graph_str_list: List of graph strings.
    :param type_dict: Dictionary to store type information.
    :return: A list of unique phrases.
    """
    seg_list = get_seg_list(graph_str_list)
    new_pairs = set()

    for seg in seg_list:
        new_seg = [item.strip() for item in seg.split(',')]
        if len(new_seg) == 1:
            _handle_single_element(new_seg, new_pairs, type_dict)
        elif len(new_seg) == 2:
            _handle_two_elements(new_seg, new_pairs, type_dict)
        elif len(new_seg) >= 3:
            _handle_three_or_more_elements(new_seg, new_pairs, type_dict)

    return list(new_pairs)


def _handle_single_element(seg, pairs, type_dict):
    pairs.add(seg[0])
    type_dict[seg[0]] = "entity"


def _handle_two_elements(seg, pairs, type_dict):
    pairs.add(f"{seg[1]} {seg[0]}")
    type_dict[f"{seg[1]} {seg[0]}"] = "attribute"
    pairs.add(seg[0])
    type_dict[seg[0]] = "entity"


def _handle_three_or_more_elements(seg, pairs, type_dict):
    if seg[1] == 'is':
        pairs.add(f"{seg[2]} {seg[0]}")
        type_dict[f"{seg[2]} {seg[0]}"] = "attribute"
    else:
        phrase = f"{seg[0]} {' '.join(seg[1:-1])} {seg[-1]}"
        pairs.add(phrase)
        type_dict[phrase] = "relation"
        pairs.add(seg[-1])
        type_dict[seg[-1]] = "entity"

    pairs.add(seg[0])
    type_dict[seg[0]] = "entity"

def encode_phrases(text_encoder, all_cand_phrases, all_ref_phrases, batch_size):
    all_encoded_phrases = _normalize_features(
        text_encoder.encode(all_cand_phrases + all_ref_phrases, batch_size=batch_size)
    )
    encoded_cands, encoded_refs = np.split(all_encoded_phrases, [len(all_cand_phrases)])
    return encoded_cands, encoded_refs


def accumulate_phrases(candidates, references):
    all_cand_phrases = []
    all_ref_phrases = []
    cand_lengths = []
    ref_lengths = []
    type_dict = {}

    for cand, refs in zip(candidates, references):
        cand_phrases = _get_graph_phrases(cand, type_dict)
        ref_phrases = _get_graph_phrases(refs, type_dict)

        all_cand_phrases.extend(cand_phrases)
        all_ref_phrases.extend(ref_phrases)

        cand_lengths.append(len(cand_phrases))
        ref_lengths.append(len(ref_phrases))

    return all_cand_phrases, all_ref_phrases, cand_lengths, ref_lengths


def compute_scores(encoded_cands, encoded_refs, cand_lengths, ref_lengths, bidirectional=False):
    scores = []
    ref_start_idx = 0

    for cand_len, ref_len in zip(cand_lengths, ref_lengths):
        cand_feats = encoded_cands[:cand_len]
        encoded_cands = encoded_cands[cand_len:]  # Update the remaining candidate features

        ref_feats = encoded_refs[ref_start_idx:ref_start_idx + ref_len]
        ref_start_idx += ref_len

        all_sims = cand_feats.dot(ref_feats.T)
        score_per_phrase = np.max(all_sims, axis=1)
        if bidirectional:
            score_per_phrase_ = np.max(all_sims, axis=0)  # recall-like
            # Normalize each score from [-1,1] to [0,1] before averaging
            norm_scores_precision = (score_per_phrase + 1) / 2
            norm_scores_recall = (score_per_phrase_ + 1) / 2
            
            precision = np.mean(norm_scores_precision)
            recall = np.mean(norm_scores_recall)
            
            # Use harmonic mean (F1-like)
            if precision + recall > 0:
                softspice_bi = 2 * (precision * recall) / (precision + recall)
            else:
                softspice_bi = 0.0
            scores.append(softspice_bi)
        else:
            scores.append(np.mean(score_per_phrase))

    return scores


def _normalize_features(features):
    """
    Normalize feature vectors.

    :param features: Feature vectors.
    :return: Normalized feature vectors.
    """
    return features / np.sqrt(np.sum(features ** 2, axis=1, keepdims=True))



from ..utils import get_seg_list, space_out_symbols_in_graph


def eval_set_match(cand, refs):
    """
    Evaluate the set match score between source and reference phrases.

    :param src_phrases: Source phrases.
    :param ref_phrases: Reference phrases.
    :return: Calculated set match score.
    """

    cand_phrases = get_seg_list(cand)
    ref_phrases = get_seg_list(refs)

    return len(cand_phrases) == len(ref_phrases) and (sorted(cand_phrases) == sorted(ref_phrases))



import os
import json

def save_results_to_json(
    args,
    candidates,
    refs,
    human_scores,
    dual_task_parse_results,
    sub_sentences_parse_dict,
    parse_dict,
    spice_score_original,
    spice_score_sub_sentences,
    spice_score_delete,
    spice_score_insert,
    spice_score_combined,
    soft_spice_score_original,
    soft_spice_score_sub_sentences,
    soft_spice_score_delete,
    soft_spice_score_insert,
    soft_spice_score_combined,
    save_path="correlation_results_caparena",
    time_tag=None,
    capture_score_subsentence=None,
    capture_score_delete=None,
    capture_score_insert=None,
    capture_score_combined=None,
):
    output_path = os.path.join(save_path, time_tag)
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    output_file = f"{output_path}/{args.dataset}_correlation_score.json"

    if not os.path.exists(output_file):
        with open(output_file, "w") as f:
            f.write("[]")

    for i in range(len(candidates)):
        cand = candidates[i]
        ref = refs[i]
        human_score = human_scores[i]
        max_len = max(len(ref), len(cand))

        # 提取所有分数
        original_spice_score = spice_score_original[i * max_len : (i + 1) * max_len]
        sub_sentences_spice_score = spice_score_sub_sentences[
            i * max_len : (i + 1) * max_len
        ]
        delete_spice_score = spice_score_delete[i * max_len : (i + 1) * max_len]
        insert_spice_score = spice_score_insert[i * max_len : (i + 1) * max_len]
        combined_spice_score = spice_score_combined[i * max_len : (i + 1) * max_len]
        original_soft_spice_score = soft_spice_score_original[
            i * max_len : (i + 1) * max_len
        ]
        sub_sentences_soft_spice_score = soft_spice_score_sub_sentences[
            i * max_len : (i + 1) * max_len
        ]
        delete_soft_spice_score = soft_spice_score_delete[
            i * max_len : (i + 1) * max_len
        ]
        insert_soft_spice_score = soft_spice_score_insert[
            i * max_len : (i + 1) * max_len
        ]
        combined_soft_spice_score = soft_spice_score_combined[
            i * max_len : (i + 1) * max_len
        ]

        # convert np.float to float
        original_spice_score = [float(x) for x in original_spice_score]
        sub_sentences_spice_score = [float(x) for x in sub_sentences_spice_score]
        delete_spice_score = [float(x) for x in delete_spice_score]
        insert_spice_score = [float(x) for x in insert_spice_score]
        combined_spice_score = [float(x) for x in combined_spice_score]
        original_soft_spice_score = [float(x) for x in original_soft_spice_score]
        sub_sentences_soft_spice_score = [
            float(x) for x in sub_sentences_soft_spice_score
        ]
        delete_soft_spice_score = [float(x) for x in delete_soft_spice_score]
        insert_soft_spice_score = [float(x) for x in insert_soft_spice_score]
        combined_soft_spice_score = [float(x) for x in combined_soft_spice_score]

        # 创建数据字典
        data = {
            "cand": cand,
            "ref": ref,
            "human_score": human_score,
            "original_spice_score": original_spice_score,
            "sub_sentences_spice_score": sub_sentences_spice_score,
            "delete_spice_score": delete_spice_score,
            "insert_spice_score": insert_spice_score,
            "combined_spice_score": combined_spice_score,
            "original_soft_spice_score": original_soft_spice_score,
            "sub_sentences_soft_spice_score": sub_sentences_soft_spice_score,
            "delete_soft_spice_score": delete_soft_spice_score,
            "insert_soft_spice_score": insert_soft_spice_score,
            "combined_soft_spice_score": combined_soft_spice_score,
        }

        # 如果启用了capture，添加相关分数
        if args.capture:
            sub_sent_capture_score = capture_score_subsentence[1][
                i * max_len : (i + 1) * max_len
            ]
            delete_capture_score = capture_score_delete[1][
                i * max_len : (i + 1) * max_len
            ]
            insert_capture_score = capture_score_insert[1][
                i * max_len : (i + 1) * max_len
            ]
            combined_capture_score = capture_score_combined[1][
                i * max_len : (i + 1) * max_len
            ]

            # convert np.float to float
            sub_sent_capture_score = [float(x) for x in sub_sent_capture_score]
            delete_capture_score = [float(x) for x in delete_capture_score]
            insert_capture_score = [float(x) for x in insert_capture_score]
            combined_capture_score = [float(x) for x in combined_capture_score]
            data.update(
                {
                    "sub_sent_capture_score": sub_sent_capture_score,
                    "delete_capture_score": delete_capture_score,
                    "insert_capture_score": insert_capture_score,
                    "combined_capture_score": combined_capture_score,
                }
            )

        with open(output_file, "r") as f:
            existing_data = json.load(f)

        existing_data.append(data)

        with open(output_file, "w") as f:
            json.dump(existing_data, f, indent=4)

    mid_res_path = os.path.join(output_path, "graph_mid_res")
    if not os.path.exists(mid_res_path):
        os.makedirs(mid_res_path)

    # save dual_task_parse_results to json
    dual_task_parse_results_file = (
        f"{mid_res_path}/{args.dataset}_dual_task_parse_results.json"
    )
    with open(dual_task_parse_results_file, "w") as f:
        json.dump(dual_task_parse_results, f, indent=4)
    print(f"Results saved to {dual_task_parse_results_file}")
    # save sub_sentences_parse_dict_object to json
    sub_sentences_parse_dict_file = (
        f"{mid_res_path}/{args.dataset}_sub_sentences_parse_dict_object.json"
    )
    with open(sub_sentences_parse_dict_file, "w") as f:
        json.dump(sub_sentences_parse_dict, f, indent=4)
    print(f"Results saved to {sub_sentences_parse_dict_file}")
    # save parse_dict to json
    parse_dict_file = f"{mid_res_path}/{args.dataset}_parse_dict.json"
    with open(parse_dict_file, "w") as f:
        json.dump(parse_dict, f, indent=4)
    print(f"Results saved to {parse_dict_file}")

    return output_file



from nltk.corpus import wordnet

from .synonym_dictionary import synonym_dictionary
from ..utils import get_seg_list

def eval_spice(cand, refs, merge_tuples_synonyms=True, synonym_match=True):
    """
    Evaluate the SPICE metric.

    :param cand_tuples: Candidate tuples from scene graph.
    :param ref_tuples: Reference tuples for comparison.
    :return: Calculated SPICE score.
    """

    cand_tuples = get_graph_tuples(cand, merge_tuples_synonyms)
    # print(refs)
    ref_tuples = get_graph_tuples(refs, merge_tuples_synonyms)
    # print(ref_tuples)
    # breakpoint()

    # print(cand_tuples)
    # print(ref_tuples)

    return calculate_spice_score(cand_tuples, ref_tuples, synonym_match)

def calculate_spice_score(cand_tuples, ref_tuples, synonym_match):
    matched_cand_indices = set()
    matched_ref_indices = set()
    total_matches = 0

    # First pass: Exact matches
    for i, cand in enumerate(cand_tuples):
        for j, ref in enumerate(ref_tuples):
            if are_tuples_match(cand, ref) and j not in matched_ref_indices:
                matched_cand_indices.add(i)
                matched_ref_indices.add(j)
                total_matches += 1
                break

    if synonym_match:
        # Second pass: WordNet-based similar matches (for unmatched candidates)
        for i, cand in enumerate(cand_tuples):
            if i not in matched_cand_indices:
                for j, ref in enumerate(ref_tuples):
                    if j not in matched_ref_indices and similar_to_any(cand, [ref]):
                        # print("Synonym match")
                        # print(cand)
                        # print(ref)
                        matched_ref_indices.add(j)
                        total_matches += 1
                        break
    # Calculate precision, recall, and F1 score
    precision = calculate_score(total_matches, len(cand_tuples))
    recall = calculate_score(total_matches, len(ref_tuples))

    assert precision <= 1 and recall <= 1, "Precision or recall is greater than 1, total_matches {0}, len(cand_tuples) {1}, len(ref_tuples) {2}".format(total_matches, len(cand_tuples), len(ref_tuples))

    return calculate_f1(precision, recall)

def similar_to_any(candidate, references):
    """
    Check if a candidate is similar to any reference tuples.

    :param candidate: The candidate tuple to compare.
    :param references: A list of reference tuples.
    :return: True if similar to any reference, False otherwise.
    """
    candidate_synsets = get_synsets(candidate)

    return any(are_tuples_match(candidate_synsets, get_synsets(ref)) for ref in references)

def get_synsets_for_word_set(word_set):
    return set().union(*[word_to_synset(word) for word in word_set])

def get_synsets(words):
    """
    Get synsets for a list of words.

    :param words: A list of words.
    :return: A set of synsets for the words.
    """
    return [get_synsets_for_word_set(word_set) for word_set in words]


def word_to_synset(word):
    """
    Process a word into its synsets.

    :param word: The word to process.
    :return: A set of synsets for the word.
    """
    word = ' '.join(word.strip().lower().split())
    lemma_synset = set()

    # If the word consists of multiple parts, join them with an underscore
    word_split = word.split()
    if len(word_split) >= 2:
        word = "_".join(word_split)

    # # Add the word itself to the synset set
    # lemma_synset.add(word)
    #
    # # Add all synsets of the word to the set
    # for sys in wordnet.synsets(word):
    #     for lemma in sys.lemmas():
    #         lemma_synset.add(lemma.name())

    lemma_synset.update(synonym_dictionary.get_synsets(word))
    lemma_synset.update(synonym_dictionary.get_stem_synsets(word))

    return lemma_synset

def are_tuples_match(synsets1, synsets2):
    """
    Determine if two lists of synsets have non-empty intersections for corresponding elements.

    :param synsets1: First list of synsets.
    :param synsets2: Second list of synsets.
    :return: True if all corresponding synsets have a non-empty intersection, False otherwise.
    """

    return len(synsets1) == len(synsets2) and all(s1.intersection(s2) for s1, s2 in zip(synsets1, synsets2))

def calculate_score(match_count, total_count):
    """
    Calculate precision or recall.

    :param match_count: The count of matched tuples.
    :param total_count: The total count of tuples.
    :return: The calculated score.
    """
    return match_count / total_count if total_count > 0 else 0

def calculate_f1(precision, recall):
    """
    Calculate the F1 score from precision and recall.

    :param precision: Precision value.
    :param recall: Recall value.
    :return: The calculated F1 score.
    """
    return 2 * precision * recall / (precision + recall) if precision + recall > 0 else 0

def get_graph_tuples(graph_str_list, merge_tuples_synonyms=True):
    """
    Get tuples from a scene graph.
    """
    seg_list = get_seg_list(graph_str_list)
    selected_obj_set = set()
    tuples = []

    for hyp in seg_list:
        lf_seg = [token.strip() for token in hyp.split(',')]
        seg_len = len(lf_seg)

        # Handle segments based on their length
        if seg_len == 1:
            add_unique_tuple(lf_seg[0], tuples, selected_obj_set)
        elif seg_len >= 2:
            process_lf_segment(lf_seg, tuples, selected_obj_set, seg_len)

    if merge_tuples_synonyms:
        return merge_tuples_based_on_synonyms(sorted(tuples, key=tuple_sort_key))
    else:
        return sorted(tuples, key=tuple_sort_key)

def tuple_sort_key(t):
    """
    Generate a sort key for a tuple of sets of strings.

    :param t: A tuple, each element of which is a set of strings.
    :return: A string that represents the sorted contents of the sets.
    """
    sorted_sets = [' '.join(sorted(s)) for s in t]  # Sort each set and join into strings
    return ' '.join(sorted_sets)  # Join the sorted strings from each set


def merge_elements_by_synsets(tuples, position, check_length, unique_sets=None):
    """
    Generalized function to merge elements within tuples based on synsets.
    """
    is_shared_set = unique_sets is not None
    unique_sets = unique_sets or []

    for t in tuples:
        if len(t) == check_length:
            merge_found = False
            for i, merged_set in enumerate(unique_sets):
                if len(get_synsets_for_word_set(t[position]).intersection(get_synsets_for_word_set(merged_set))) > 0:
                    merge_found = True
                    unique_sets[i] = merged_set.union(t[position])
                    break
            if not merge_found and not is_shared_set:
                unique_sets.append(t[position])

    for t in tuples:
        if len(t) == check_length:
            for merged_set in unique_sets:
                if len(get_synsets_for_word_set(t[position]).intersection(get_synsets_for_word_set(merged_set))) > 0:
                    t[position].update(merged_set)
                    break

    return unique_sets if not is_shared_set else None


def merge_tuples_based_on_synonyms(tuples):
    """
    Merge nodes, attributes, and relations in tuples based on synonyms.
    """
    # Create a shared unique set for specific merge operations
    shared_unique_set = merge_elements_by_synsets(tuples, 0, 1)  # Merging nodes and initializing shared set

    # Use the shared set for merging nodes in three-element tuples and at the end of three-element tuples
    merge_elements_by_synsets(tuples, 0, 3, shared_unique_set)
    merge_elements_by_synsets(tuples, 2, 3, shared_unique_set)

    # Merging attributes in two-element tuples
    merge_elements_by_synsets(tuples, 1, 2)

    # Merging relations in three-element tuples
    merge_elements_by_synsets(tuples, 1, 3)

    merged_tuples = merge_tuples(tuples)

    return merged_tuples

def merge_tuples(tuples):
    """
    Merge tuples if they have similar elements.
    """
    merged_tuples = []
    for t in tuples:
        merge_found = False
        for i, mt in enumerate(merged_tuples):
            if similar_to_any(t, [mt]):
                merged_tuples[i] = merge_two_tuples(t, mt)
                merge_found = True
                break
        if not merge_found:
            merged_tuples.append(t)

    return merged_tuples

def merge_two_tuples(tuple1, tuple2):
    """
    Merge two tuples that have synonyms in common.

    :param tuple1: A tuple.
    :param tuple2: Another tuple.
    :return: A merged tuple.
    """
    # print(tuple1)
    # print(tuple2)
    return [t1.union(t2) for t1, t2 in zip(tuple1, tuple2)]

def add_unique_tuple(item, tuples, selected_obj_set):
    """
    Adds a unique tuple from an item.
    """
    if item not in selected_obj_set:
        tuples.append([{item}])
        selected_obj_set.add(item)


def process_lf_segment(lf_seg, tuples, selected_obj_set, seg_len):
    """
    Processes a segment of length 2 or more and adds appropriate tuples.
    """
    # Construct the tuple string based on segment length
    if seg_len == 2 or (seg_len == 3 and lf_seg[1] == 'is'):
        tuple_str = lf_seg[0] + ' ' + lf_seg[-1]
        if tuple_str not in selected_obj_set:
            tuples.append(({lf_seg[0]}, {lf_seg[-1]}))
            selected_obj_set.add(tuple_str)
        add_unique_tuple(lf_seg[0], tuples, selected_obj_set)

    elif seg_len == 3:
        tuple_str = ' '.join(lf_seg)
        if tuple_str not in selected_obj_set:
            tuples.append(({lf_seg[0]}, {lf_seg[1]}, {lf_seg[2]}))
            selected_obj_set.add(tuple_str)
        add_unique_tuple(lf_seg[0], tuples, selected_obj_set)
        add_unique_tuple(lf_seg[2], tuples, selected_obj_set)

    elif seg_len > 3:
        tuple_str = lf_seg[0] + ' ' + ' '.join(lf_seg[1:-1]) + ' ' + lf_seg[-1]
        if tuple_str not in selected_obj_set:
            tuples.append(({lf_seg[0]}, {" ".join(lf_seg[1:-1])}, {lf_seg[-1]}))
            selected_obj_set.add(tuple_str)
        add_unique_tuple(lf_seg[0], tuples, selected_obj_set)
        add_unique_tuple(lf_seg[-1], tuples, selected_obj_set)






import scipy

from typing import Dict, List, Any, Tuple
from disco_sg_camera.src.factual_scene_graph.evaluation.evaluator import Evaluator

def print_correlation(avg_score, correlation_c, correlation_b, pearson_correlation, score_name):
    """Print the correlation results."""
    print(f"{score_name} score: {avg_score:.3f}")
    print(f"{score_name} Tau-c: {100 * correlation_c:.3f}")
    print(f"{score_name} Tau-b: {100 * correlation_b:.3f}")
    print(f"{score_name} Pearson correlation: {100 * pearson_correlation:.3f}")

def compute_correlation(scores, human_scores, score_name):
    """Compute all kendall and pearson correlation between model scores and human scores."""
    avg_score = sum(scores) / len(scores)
    correlation_c = scipy.stats.kendalltau(scores, human_scores, variant="c")[0]
    correlation_b = scipy.stats.kendalltau(scores, human_scores, variant="b")[0]
    pearson_correlation = scipy.stats.pearsonr(scores, human_scores)[0]
    print_correlation(avg_score, correlation_c, correlation_b, pearson_correlation, score_name)
    return avg_score, correlation_c, correlation_b, pearson_correlation

def print_original_metrics(spice_score_original, soft_spice_score_original, human_scores_flat):
    print("---- Original SPICE score ----")
    spice_score_original, Tau_c_ori, Tau_b_ori, Pearson_ori = compute_correlation(spice_score_original, human_scores_flat, score_name="SPICE score original")
    print("---- Original Soft SPICE score ----")
    soft_spice_score_original, Tau_c_ori_soft, Tau_b_ori_soft, Pearson_ori_soft = compute_correlation(soft_spice_score_original, human_scores_flat, score_name="Soft SPICE score original")

def print_sub_sentences_metrics(spice_score_sub_sentences, soft_spice_score_sub_sentences, human_scores_flat):
    print("---- Sub-sentences SPICE score ----")
    spice_score_sub_sentences, Tau_c_sub_sentences, Tau_b_sub_sentences, Pearson_sub_sentences = compute_correlation(spice_score_sub_sentences, human_scores_flat, score_name="SPICE score sub-sentences")
    print("---- Sub-sentences Soft SPICE score ----")
    soft_spice_score_sub_sentences, Tau_c_sub_sentences_soft, Tau_b_sub_sentences_soft, Pearson_sub_sentences_soft = compute_correlation(soft_spice_score_sub_sentences, human_scores_flat, score_name="Soft SPICE score sub-sentences")

def print_three_task_metrics(spice_score_delete, soft_spice_score_delete, spice_score_insert, soft_spice_score_insert, spice_score_combined, soft_spice_score_combined, human_scores_flat):
    print("---- Delete SPICE score ----")
    spice_score_delete, Tau_c_delete, Tau_b_delete, Pearson_delete = compute_correlation(spice_score_delete, human_scores_flat, score_name="SPICE score for delete task")
    print("---- Delete Soft SPICE score ----")
    soft_spice_score_delete, Tau_c_delete_soft, Tau_b_delete_soft, Pearson_delete_soft = compute_correlation(soft_spice_score_delete, human_scores_flat, score_name="Soft SPICE score for delete task")
    print("---- Insert SPICE score ----")
    spice_score_insert, Tau_c_insert, Tau_b_insert, Pearson_insert = compute_correlation(spice_score_insert, human_scores_flat, score_name="SPICE score for insert task")
    print("---- Insert Soft SPICE score ----")
    soft_spice_score_insert, Tau_c_insert_soft, Tau_b_insert_soft, Pearson_insert_soft = compute_correlation(soft_spice_score_insert, human_scores_flat, score_name="Soft SPICE score for insert task")
    print("---- Combined SPICE score ----")
    spice_score_combined, Tau_c_combined, Tau_b_combined, Pearson_combined = compute_correlation(spice_score_combined, human_scores_flat, score_name="SPICE score for combined task")
    print("---- Combined Soft SPICE score ----")
    soft_spice_score_combined, Tau_c_combined_soft, Tau_b_combined_soft, Pearson_combined_soft = compute_correlation(soft_spice_score_combined, human_scores_flat, score_name="Soft SPICE score for combined task")

def print_original_metrics_capture(capture_score_original, human_scores_flat):
    print("---- Original CAPTURE score ----")
    capture_score_original, Tau_c_ori, Tau_b_ori, Pearson_ori = compute_correlation(capture_score_original[1], human_scores_flat, score_name="CAPTURE score original")

def print_sub_sentences_metrics_capture(capture_score_sub_sentences, human_scores_flat):
    print("---- Sub-sentences CAPTURE score ----")
    capture_score_sub_sentences, Tau_c_sub_sentences, Tau_b_sub_sentences, Pearson_sub_sentences = compute_correlation(capture_score_sub_sentences[1], human_scores_flat, score_name="CAPTURE score sub-sentences")

def print_three_task_metrics_capture(capture_score_delete, capture_score_insert, capture_score_combined, human_scores_flat):
    print("---- Delete CAPTURE score ----")
    capture_score_delete, Tau_c_delete, Tau_b_delete, Pearson_delete = compute_correlation(capture_score_delete[1], human_scores_flat, score_name="CAPTURE score for delete task")
    print("---- Insert CAPTURE score ----")
    capture_score_insert, Tau_c_insert, Tau_b_insert, Pearson_insert = compute_correlation(capture_score_insert[1], human_scores_flat, score_name="CAPTURE score for insert task")
    print("---- Combined CAPTURE score ----")
    capture_score_combined, Tau_c_combined, Tau_b_combined, Pearson_combined = compute_correlation(capture_score_combined[1], human_scores_flat, score_name="CAPTURE score for combined task")

def evaluate_graphs(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict: Dict[str, Any],
    evaluator: Evaluator,
    return_graphs: bool,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""

    cand_graphs = [
        (
            [parse_dict[cand] for cand in cand.values()]
            if isinstance(cand, dict)
            else [parse_dict[cand]]
        )
        for cand in candidates
    ]
    ref_graphs = [
        [parse_dict[ref] for ref in ref.values()] if isinstance(ref, dict) else [ref]
        for ref in refs
    ]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    length_diff_times = len(cand_graphs) // len(ref_graphs)
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(length_diff_times)]
    assert len(cand_graphs) == len(ref_graphs), "Candidate and reference lengths do not match"

    spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
    )

    soft_spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="soft_spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
        bidirectional=True,
    )

    return spice_res, soft_spice_res

def evaluate_graphs_flickr8k(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict: Dict[str, Any],
    evaluator: Evaluator,
    return_graphs: bool,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    cand_graphs = [
        [parse_dict[cand] for cand in cand.values()] if isinstance(cand, dict) else [parse_dict[cand]] for cand in candidates
    ]
    ref_graphs = [[parse_dict[ref_i] for ref_i in ref] for ref in refs]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(3)]

    spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
    )

    soft_spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="soft_spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
        bidirectional=True,
    )

    return spice_res, soft_spice_res

def evaluate_graphs_sft(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict: Dict[str, Any],
    evaluator: Evaluator,
    return_graphs: bool,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    cand_graphs = [
        (
            [parse_dict[cand] for cand in cand.values()]
            if isinstance(cand, dict)
            else [cand]
        )
        for cand in candidates
    ]
    ref_graphs = [
        [parse_dict[ref] for ref in ref.values()] if isinstance(ref, dict) else [ref]
        for ref in refs
    ]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(3)]

    spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
    )

    soft_spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="soft_spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
        bidirectional=True,
    )

    return spice_res, soft_spice_res

def evaluate_graphs_caparena(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict: Dict[str, Any],
    evaluator: Evaluator,
    return_graphs: bool,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    cand_graphs = [
        (
            [parse_dict[cand] for cand in cand.values()]
            if isinstance(cand, dict)
            else [parse_dict[cand]]
        )
        for cand in candidates
    ]
    ref_graphs = [
        [parse_dict[ref] for ref in ref.values()] if isinstance(ref, dict) else [ref]
        for ref in refs
    ]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(2)]

    spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
    )

    soft_spice_res = evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="soft_spice",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
        bidirectional=True,
    )

    return spice_res, soft_spice_res

def evaluate_graphs_capture(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict_capture: Dict[str, Any],
    capture_evaluator: Evaluator,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    cand_graphs = [
        (
            [parse_dict_capture[cand] for cand in cand.values()]
            if isinstance(cand, dict)
            else [parse_dict_capture[cand]]
        )
        for cand in candidates
    ]
    ref_graphs = [
        (
            [parse_dict_capture[ref] for ref in ref.values()]
            if isinstance(ref, dict)
            else [ref]
        )
        for ref in refs
    ]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(3)]

    assert len(cand_graphs) == len(
        ref_graphs
    ), "Candidate and reference lengths do not match"
    refs_example_dict = {}
    cands_example_dict = {}
    for i in range(len(ref_graphs)):
        example_id = f"example_{i}"
        refs_example_dict[example_id] = ref_graphs[i]
        cands_example_dict[example_id] = cand_graphs[i]

    capture_score = capture_evaluator.compute_score_post(
        refs_example_dict,
        cands_example_dict,
    )

    return capture_score

def evaluate_graphs_ori(candidates, refs, parse_dict, evaluator, return_graphs):
    """Evaluate the graphs and return the results."""
    cand_graphs = [parse_dict[cand] for cand in candidates]
    ref_graphs = [[parse_dict[ref_i] for ref_i in ref] for ref in refs]
    spice_out = evaluator.evaluate(cand_graphs, ref_graphs, method='spice',
                              beam_size=1, batch_size=64, max_input_len=512,
                              max_output_len=512, return_graphs=return_graphs)
    soft_spice_out = evaluator.evaluate(cand_graphs, ref_graphs, method='soft_spice',
                                beam_size=1, batch_size=64, max_input_len=512,
                                max_output_len=512, return_graphs=return_graphs, bidirectional=True,)
    return spice_out, soft_spice_out

def evaluate_graphs_capture_flicke8k(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict_capture: Dict[str, Any],
    capture_evaluator: Evaluator,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    cand_graphs = [
        (
            [parse_dict_capture[cand] for cand in cand.values()]
            if isinstance(cand, dict)
            else [parse_dict_capture[cand]]
        )
        for cand in candidates
    ]
    ref_graphs = [
        (
            [parse_dict_capture[ref] for ref in ref.values()]
            if isinstance(ref, dict)
            else [parse_dict_capture[r] for r in ref]
        )
        for ref in refs
    ]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(3)]

    assert len(cand_graphs) == len(
        ref_graphs
    ), "Candidate and reference lengths do not match"
    refs_example_dict = {}
    cands_example_dict = {}
    for i in range(len(ref_graphs)):
        example_id = f"example_{i}"
        refs_example_dict[example_id] = ref_graphs[i]
        cands_example_dict[example_id] = cand_graphs[i]

    capture_score = capture_evaluator.compute_score_post(
        refs_example_dict,
        cands_example_dict,
    )

    return capture_score

def evaluate_graphs_capture_caparena(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict_capture: Dict[str, Any],
    capture_evaluator: Evaluator,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    cand_graphs = [
        (
            [parse_dict_capture[cand] for cand in cand.values()]
            if isinstance(cand, dict)
            else [parse_dict_capture[cand]]
        )
        for cand in candidates
    ]
    ref_graphs = [
        (
            [parse_dict_capture[ref] for ref in ref.values()]
            if isinstance(ref, dict)
            else [ref]
        )
        for ref in refs
    ]

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    if len(cand_graphs) != len(ref_graphs):
        ref_graphs = [item for item in ref_graphs for _ in range(2)]

    assert len(cand_graphs) == len(
        ref_graphs
    ), "Candidate and reference lengths do not match"
    refs_example_dict = {}
    cands_example_dict = {}
    for i in range(len(ref_graphs)):
        example_id = f"example_{i}"
        refs_example_dict[example_id] = ref_graphs[i]
        cands_example_dict[example_id] = cand_graphs[i]

    capture_score = capture_evaluator.compute_score_post(
        refs_example_dict,
        cands_example_dict,
    )

    return capture_score

def set_match_evaluate_graphs(
    candidates: List[Dict[str, str]],
    refs: List[str],
    parse_dict: Dict[str, Any],
    evaluator: Evaluator,
    return_graphs: bool,
) -> Tuple[List[float], Any, Any]:
    """Enhanced graph evaluation."""
    # check if candidates is a list of dicts
    if isinstance(candidates[0], dict):
        cand_graphs = [
            [parse_dict[value] for value in cand.values()] for cand in candidates
        ]
        ref_graphs = [[parse_dict[ref]] for ref in refs]
    else:
        cand_graphs = [[parse_dict[cand]] for cand in candidates]
        ref_graphs = [[ref] for ref in refs]

    # Normalize lengths
    # check if cand_graphs is a list of lists
    if isinstance(candidates[0], dict):
        for i in range(len(cand_graphs)):
            if len(cand_graphs[i]) != 3:
                print(
                    f"Warning: Candidate {i} has {len(cand_graphs[i])} graphs instead of 3"
                )
                continue
            ref_graphs[i] = ref_graphs[i] * 3
    else:
        ref_graphs = ref_graphs  # do nothing
        assert len(cand_graphs) == len(
            ref_graphs
        ), "Candidate and reference lengths do not match"

    # Flatten lists
    cand_graphs = [item for sublist in cand_graphs for item in sublist]
    ref_graphs = [[item] for sublist in ref_graphs for item in sublist]

    return evaluator.evaluate(
        cand_graphs,
        ref_graphs,
        method="set_match",
        # num_beams=1,
        batch_size=32,
        max_input_len=512,
        max_output_len=512,
        return_graphs=return_graphs,
        # num_return_sequences=5,
        # temperature=0.5,
        # top_k=50,
        # top_p=0.95,
        do_sample=False,
    )



import os
import json
import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Dict, List
from huggingface_hub import hf_hub_download

class T5ForDualTask(nn.Module):
    def __init__(
        self,
        model_name: str = "google/flan-t5-base",
        use_lora: bool = False,
        lora_r: int = 8,
        lora_alpha: int = 32,
        lora_dropout: float = 0.1,
        lora_target_modules: List[str] = ["q", "v"],
        device = "cpu",
    ):
        super().__init__()
        from transformers import T5ForConditionalGeneration
        self.model = T5ForConditionalGeneration.from_pretrained(model_name).to(device)
        self.use_lora = use_lora

        self.classifier = nn.Linear(
            self.model.config.d_model, 2
        )  # DELETE(1) and KEEP(0)

        try:
            # Attempt to load the classifier weights from the Hugging Face Hub
            classifier_path = hf_hub_download(
                repo_id=model_name,
                filename="classifier.pt"
            )
            self.classifier.load_state_dict(
                torch.load(classifier_path, weights_only=True)
            )
            self.classifier.to(device)
            self.classifier.eval()  # Set to eval mode
            print("Successfully loaded classifier weights from Hugging Face Hub.")
        except Exception as e:
            print(f"Error download/load classifier weights: {e}")

        if use_lora:
            from peft import get_peft_model, LoraConfig, TaskType
            lora_config = LoraConfig(
                task_type=TaskType.SEQ_2_SEQ_LM,
                r=lora_r,
                lora_alpha=lora_alpha,
                lora_dropout=lora_dropout,
                target_modules=lora_target_modules,
            )
            self.model = get_peft_model(self.model, lora_config)

    def forward_delete(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: torch.Tensor = None,
        span_mask: torch.Tensor = None,
        triple_spans: torch.Tensor = None,
    ) -> Dict[str, torch.Tensor]:
        encoder_outputs = self.model.encoder(
            input_ids=input_ids, attention_mask=attention_mask, return_dict=True
        )
        cls_hidden_state = encoder_outputs.last_hidden_state

        all_logits = []
        all_labels = []
        for i in range(input_ids.size(0)):
            span = triple_spans[i]  # list of (start, end) tuples for sample i
            masked_span = span[span_mask[i]]
            triple_reps = []
            for start, end in masked_span:
                # Obtain the embedding for the triple span by averaging token embeddings.
                span_emb = cls_hidden_state[i, start:end+1, :].mean(dim=0)  # TODO: try some new methods
                triple_reps.append(span_emb)
            # Stack the triples representations and classify each.
            if not triple_reps:
                continue
            triple_reps = torch.stack(triple_reps, dim=0)
            logits = self.classifier(triple_reps)
            all_logits.append(logits)
            if labels is not None:
                masked_labels = labels[i][span_mask[i]]
                all_labels.append(masked_labels)

        all_logits = torch.cat(all_logits, dim=0)
        outputs = {"logits": all_logits}

        if labels is not None:
            all_labels = torch.cat(all_labels, dim=0)
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(all_logits, all_labels)
            outputs["loss"] = loss

        return outputs
    
    def forward_delete_maskBIO(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: torch.Tensor = None,
        span_mask: torch.Tensor = None,
        triple_spans: torch.Tensor = None,
    ) -> Dict[str, torch.Tensor]:
        encoder_outputs = self.model.encoder(
            input_ids=input_ids, attention_mask=attention_mask, return_dict=True
        )
        cls_hidden_state = encoder_outputs.last_hidden_state

        B, L, H = cls_hidden_state.size()

        batch_idx, start_idx, end_idx, span_labels = [], [], [], []

        for b in range(B):
            keep = span_mask[b].nonzero(as_tuple=False).squeeze(1)
            if keep.numel() == 0:
                continue
            spans = triple_spans[b].detach().clone()[keep]
            start, end = spans[:, 0], spans[:, 1]
            k = start.size(0)

            batch_idx.append(torch.full((k,), b, device=cls_hidden_state.device))
            start_idx.append(start)
            end_idx.append(end)

            if labels is not None:
                span_labels.append(labels[b][keep])
            
        if not batch_idx:
            empty_tensor = torch.zeros((0, 2), device=cls_hidden_state.device)
            return {"logits": empty_tensor}
        
        batch_idx = torch.cat(batch_idx, dim=0)
        start_idx = torch.cat(start_idx, dim=0)
        end_idx = torch.cat(end_idx, dim=0)
        if labels is not None:
            span_labels = torch.cat(span_labels, dim=0)
        
        cumsum = torch.cumsum(cls_hidden_state, dim=1)
        cumsum = F.pad(cumsum, (0, 0, 1, 0))

        span_sum = cumsum[batch_idx, end_idx + 1] - cumsum[batch_idx, start_idx]
        span_len = (end_idx - start_idx + 1).unsqueeze(1)
        span_repr = span_sum / span_len

        logits = self.classifier(span_repr)
        outputs = {"logits": logits}
        if labels is not None:
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(logits.view(-1, 2), span_labels.view(-1))
            outputs["loss"] = loss
        return outputs

    def forward_insert(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        decoder_input_ids: torch.Tensor,
        labels: torch.Tensor = None,
    ) -> Dict[str, torch.Tensor]:
        outputs = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            labels=labels,
            return_dict=True,
        )

        return {"loss": outputs.loss, "logits": outputs.logits}

    def generate(
        self, input_ids: torch.Tensor, attention_mask: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        return self.model.generate(
            input_ids=input_ids, attention_mask=attention_mask, **kwargs
        )


# def main():
#     # Example usage
#     model = T5ForDualTask(model_name="sqlinn/DiscoSG-Refiner-Large-t5-only", use_lora=False)
#     input_ids = torch.randint(0, 1000, (2, 10))  # Dummy input
#     attention_mask = torch.ones((2, 10), dtype=torch.long)  # Dummy attention mask
#     labels = torch.randint(0, 2, (2, 10))  # Dummy labels
#     span_mask = torch.tensor([[1, 0, 1, 0, 1, 0, 1, 0, 1, 0], [1, 1, 0, 0, 1, 1, 0, 0, 1, 1]])
#     triple_spans = torch.tensor([[(0, 2), (4, 6)], [(0, 3), (4, 5)]])

#     outputs = model.forward_delete(
#         input_ids=input_ids,
#         attention_mask=attention_mask,
#         labels=labels,
#         span_mask=span_mask,
#         triple_spans=triple_spans,
#     )
#     print(outputs)

# if __name__ == "__main__":
#     main()


import functools
import re

import tabulate

def seed_everything(seed=42):
    """
    设置所有相关库的随机种子，确保实验可重复性
    
    参数:
        seed (int): 随机种子值，默认为42
    """
    import random
    import numpy as np
    import torch
    import os
    
    random.seed(seed)
    
    np.random.seed(seed)
    
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    
    os.environ['PYTHONHASHSEED'] = str(seed)
    
    print(f"Random seed set as {seed}")


def get_seg_list(graphs):
    """
    Extracts segments from a single graph or a list of graphs.

    :param graphs: A single graph string or a list of graph strings.
    :return: A list of unique segments from the graph(s).
    """

    def extract_segments(graph):
        """
        Extract segments from an individual graph string.

        :param graph: A single graph string.
        :return: A list of segments from the graph.
        """
        formatted_graph = space_out_symbols_in_graph(graph)
        return [segment.strip('()').strip() for segment in formatted_graph.split(') , (')]

    if isinstance(graphs, str):
        segments = extract_segments(graphs)
    elif isinstance(graphs, list):
        segments = [seg for graph in graphs for seg in extract_segments(graph)]
    else:
        raise TypeError('Input must be a string or a list of strings')

    # remove the duplicates, note this might influence the evaluation performance
    return list(set(segments))


def space_out_symbols_in_graph(graph_str):
    # Add spaces around parentheses and commas, then split into words
    formatted_str = graph_str.replace('(', ' ( ').replace(')', ' ) ').replace(',', ' , ')

    # Use strip to remove leading/trailing whitespace and join the words back into a string
    return ' '.join(word.strip() for word in formatted_str.split())


def is_graph_format(input_string):
    """
    Check if the input string follows the graph format.

    :param input_string: A string to check.
    :return: True if the string contains elements in parentheses, False otherwise.
    """
    # Pattern to match any content within parentheses
    graph_pattern = r"\(.*?\)"

    return bool(re.search(graph_pattern, input_string))

def remove_factual_chars(text):
    """
    Remove specific substrings from the text, including patterns like ':digit'.

    :param text: Input text from which substrings will be removed.
    :return: Text after removing specific substrings.
    """
    # Direct string replacements
    replacements = ['v:', 'pv:']
    for replacement in replacements:
        text = text.replace(replacement, '')

    # Using regular expression to remove patterns like ':digit'
    text = re.sub(r':\d+', '', text)

    return text

def clean_graph_string(fact_str):
    # Split the string into individual facts
    facts = fact_str.strip().split(') ,')
    # remove truncated parentheses
    if not fact_str.endswith(')') and len(facts) > 1:
        facts = facts[:-1]
    elif not fact_str.endswith(')') and len(facts) == 1:
        facts = [facts[0].split(',')[0]]
    # Use a set to filter out duplicate facts
    unique_facts = set()
    for fact in facts:
        fact = fact.strip().strip('()').strip()
        if fact:
            unique_facts.add(fact)

    # sort unique_facts

    unique_facts = sorted(unique_facts)

    # Reconstruct the string with unique facts
    unique_fact_str = ' , '.join([f'( {fact} )' for fact in unique_facts])
    return unique_fact_str


def tprint(graph, file=None):
    """
    Print a scene graph as a table.
    The printed strings contain essential information about the parsed scene graph.
    """
    assert isinstance(graph, dict), 'Input must be a dictionary'
    _print = functools.partial(print, file=file)

    _print('Entities:')
    entities_data = [
        [e['head'].lower(), e.get('quantity', ''), ','.join(e.get('attributes', set()))]
        for e in graph['entities']
    ]
    _print(tabulate.tabulate(entities_data, headers=['Entity', 'Quantity', 'Attributes'], tablefmt=_tabulate_format))

    _print('Relations:')
    relations_data = [
        [
            graph['entities'][rel['subject']]['head'].lower(),
            rel['relation'].lower(),
            graph['entities'][rel['object']]['head'].lower()
        ]
        for rel in graph['relations']
    ]
    _print(tabulate.tabulate(relations_data, headers=['Subject', 'Relation', 'Object'], tablefmt=_tabulate_format))


_tabulate_format = tabulate.TableFormat(
    lineabove=tabulate.Line("+", "-", "+", "+"),
    linebelowheader=tabulate.Line("|", "-", "+", "|"),
    linebetweenrows=None,
    linebelow=tabulate.Line("+", "-", "+", "+"),
    headerrow=tabulate.DataRow("|", "|", "|"),
    datarow=tabulate.DataRow("|", "|", "|"),
    padding=1, with_header_hide=None
)






from setuptools import setup, find_packages

setup(
    name='discosg',
    version='0.0.4',
    author='Shaoqing Lin, Zhuang Li',
    author_email='sqlinn@whu.edu.cn',
    description='A package for discourse-level scene graph parsing and evaluation.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/ShaoqLin/DiscoSG',
    package_dir={'': "src"},
    packages=find_packages(where='src'),
    include_package_data=True,
    python_requires=">=3.8.0",
    install_requires=[
        'torch<=2.6.0',
        'transformers',
        'tqdm',
        'nltk',
        'tqdm',
        'peft',
        'huggingface_hub'
        # Add other dependencies needed for your package
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Add additional classifiers as appropriate for your project
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: OS Independent',
    ],
)



import json
from typing import List, Dict, Any

def load_parsed_captions(
    file_path: str,
) -> Dict[str, Any]:
    """Load parsed captions from a file."""
    if file_path.endswith(".json"):
        print(f"Loading parsed captions from {file_path}")
        with open(file_path, "r") as f:
            parsed_captions = json.load(f)
    else:
        raise ValueError(f"Unsupported file format: {file_path}")
    return parsed_captions

def parse_captions(
    captions: List[str], parser, sub_sentence: bool = False, batch_size: int = 32, num_beams: int = 3
) -> Dict[str, Any]:
    """Parse captions with enhanced processing."""
    parse_results = parser.parse(
        captions,
        batch_size=batch_size,
        max_input_len=512 if sub_sentence else 256,
        max_output_len=512 if sub_sentence else 256,
        num_beams=num_beams,
        return_text=True,
        sub_sentence=sub_sentence,
        num_return_sequences=5,
        temperature=0.5,
        top_k=50,
        top_p=0.95,
        do_sample=True,
    )
    return {caption: result for caption, result in zip(captions, parse_results)}

def parse_captions_fix(
    descriptions: List[str],
    graph_to_fix,
    parser,
    skip_toolong,
    skip_len,
    max_input_len: int = 2048,
    max_output_len: int = 512,
    max_triples_num: int = 256,
    task: str = "insert_delete",
    batch_size: int = 32,
    num_beams: int = 1,
    do_sample: bool = False,
    top_k: int = 50,
    top_p: float = 1.0,
    temperature: float = 1.0,
) -> Dict[str, Any]:
    """Parse captions with enhanced processing."""
    parse_results = parser.parse(
        descriptions=descriptions,
        graph_to_fix=graph_to_fix,
        batch_size=batch_size,
        task=task,
        max_input_len=max_input_len,
        max_output_len=max_output_len,
        max_triples_num=max_triples_num,
        skip_toolong=skip_toolong,
        skip_len=skip_len,
        num_beams=num_beams,
        do_sample=do_sample,
        top_k=top_k,
        top_p=top_p,
        temperature=temperature,
    )
    if task == "insert" or task == "delete":
        return {caption: result for caption, result in zip(descriptions, parse_results)}
    elif task == "insert_delete":
        delete_results = parse_results["delete"]
        insert_results = parse_results["insert"]
        return {"delete": delete_results, "insert": insert_results}
    elif task == "delete_before_insert":
        delete_results = parse_results["delete"]
        insert_results = parse_results["insert"]
        return {"delete": delete_results, "insert": insert_results}
    else:
        raise ValueError(f"Unsupported task: {task}")




import re
from typing import List, Dict


def remove_space(text: str) -> str:
    """Remove unnecessary spaces in text."""
    text = text.replace(" ,", ",").replace(", ", ",")
    text = text.replace(" (", "(").replace("( ", "(")
    text = text.replace(" )", ")").replace(") ", ")")
    return text

def normalize_triple(triple_str: str) -> List[str]:
    """Normalize a triple string into a list."""
    triple = triple_str.strip("()").split(",")
    return [part.strip() for part in triple]

def extract_triples(scene_graph: str) -> List[List[str]]:
    """Extract triples from scene graph string."""
    scene_graph = remove_space(scene_graph)
    raw_triples = scene_graph.split("),(")
    processed = []
    for triple in raw_triples:
        clean_triple = triple.strip().strip("()")
        if clean_triple:
            normalized = normalize_triple(clean_triple)
            if len(normalized) == 3:
                processed.append(normalized)
    return processed

def compare_triples(triple1: List[str], triple2: List[str]) -> bool:
    """Compare two triples regardless of order."""
    return sorted(triple1) == sorted(triple2)

def count_matching_triples(
    predictions: List[List[str]], ground_truths: List[List[str]]
) -> int:
    """Count matching triples between predictions and ground truths."""
    count = 0
    for pred in predictions:
        for gt in ground_truths:
            if compare_triples(pred, gt):
                count += 1
                break
    return count

def clean_text(text: str) -> str:
    """Clean and normalize text."""
    text = re.sub(r'"+', '"', text)
    text = text.replace('"', "'")
    text = re.sub(r"\n+", " ", text)
    return text

def clean_text_caparena(text: str) -> str:
    """Clean and normalize text."""
    # strip
    text = text.strip()
    text = re.sub(r'"+', '"', text)
    text = text.replace('"', "'")
    text = re.sub(r"\n+", " ", text)
    text = re.sub(r"\t+", "", text)
    # text = re.sub("*", "", text)
    text = text.replace("*", "")
    assert "*" not in text, f"Error: {text}"
    assert "\n" not in text, f"Error: {text}"
    assert "\t" not in text, f"Error: {text}"
    return text

def merge_delete_insert_results(
    insert_res: str, delete_res: str
) -> List[List[str]]:
    """Merge delete and insert results into a single list of triples."""
    insert_res_list = extract_triples(insert_res)
    delete_res_list = extract_triples(delete_res)
    merged_res_list = insert_res_list + delete_res_list
    merged_res_list = list(set([tuple(triple) for triple in merged_res_list]))
    merged_res_list_str = " , ".join(
        [f"( {' , '.join(triple)} )" for triple in merged_res_list]
    )
    return merged_res_list, merged_res_list_str

def get_or_create_entity_index(entity_name, graph, entity_map):
        if entity_name not in entity_map:
            new_index = len(graph['entities'])
            graph['entities'].append({'head': entity_name, 'quantity': '', 'attributes': set()})
            entity_map[entity_name] = new_index
        else:
            new_index = entity_map[entity_name]

        return new_index

def graph_string_to_object(graph_text):
    graph = {'entities': [], 'relations': []}
    entity_map = {}  # Entity name to index mapping

    # Process each relation in the description
    relation_strs = graph_text.strip().split(') , ')
    for relation_str in relation_strs:
        relation_str = relation_str.strip().strip('()')
        parts = [part.strip() for part in relation_str.split(',')]

        if len(parts) != 3 and len(relation_strs) > 1:
            continue  # Skip malformed relations
        elif len(parts) != 3 and len(relation_strs) == 1:
            get_or_create_entity_index(parts[0], graph, entity_map)
        else:
            subject, relationship, object_ = parts

            subject_index = get_or_create_entity_index(subject, graph, entity_map)

            if relationship == 'is':
                if object_.isdigit():  # Quantity
                    graph['entities'][subject_index]['quantity'] = object_
                else:  # Attribute
                    graph['entities'][subject_index]['attributes'].add(object_)
            else:
                object_index = get_or_create_entity_index(object_, graph, entity_map)
                # Add relation
                graph['relations'].append({'subject': subject_index, 'relation': relationship, 'object': object_index})

    return graph

def collect_unique_captions_merge_refs_into_one_sentence(
    candidates: List[Dict[str, str]], refs: List[str]
) -> List[str]:
    """Collect unique captions from candidates and references."""
    caption_set = set()
    # merged_refs = [' '.join(ref_list) for ref_list in refs]
    # make sure each ref ends with a period during merging
    # for ref in refs:
    #     if ref and not ref.endwith('.'):
    #         ref += '.'
    for ref_list in refs:
        for ref in ref_list:
            if ref and not ref.endswith('.'):
                ref += '.'
    merged_refs = [' '.join(ref_list) for ref_list in refs]
    caption_set.update(merged_refs)
    return list(caption_set)

def collect_unique_captions(candidates, refs):
    """Collect unique captions from candidates and references."""
    caption_set = set(candidates)  # Add all candidates to the set
    for ref_list in refs:
        caption_set.update(ref_list)  # Add all elements from each list in refs
    return list(caption_set)


import json
import numpy as np

from datasets import load_dataset
from tqdm import tqdm
from typing import Dict, List
from disco_sg_camera.src.factual_scene_graph.triple_utils import (
    clean_text,
    clean_text_caparena,
    extract_triples,
)

def collect_unique_captions(
    candidates: List[Dict[str, str]], refs: List[str]
) -> List[str]:
    """Collect unique captions from candidates and references."""
    caption_set = set()
    for cand in candidates:
        # check if cand is a dict
        if isinstance(cand, dict):
            for value in cand.values():
                caption_set.add(value)
        else:
            caption_set.add(cand)
    for ref in refs:
        # check if ref is a dict
        if isinstance(ref, dict):
            for value in ref.values():
                caption_set.add(value)
        else:
            caption_set.add(ref)
    return list(caption_set)

def collect_unique_captions_flickr8k(candidates, refs):
    """Collect unique captions from candidates and references."""
    caption_set = set(candidates)  # Add all candidates to the set
    for ref_list in refs:
        caption_set.update(ref_list)  # Add all elements from each list in refs
    return list(caption_set)

def load_detailcaps_dataset(
    dataset_name_or_path="foundation-multimodal-models/DetailCaps-4870", split="test"
):
    dataset = load_dataset(dataset_name_or_path, split=split)

    refs = []
    candidates = []
    human_scores = []

    # Process dataset
    for data in tqdm(dataset, desc="Processing dataset"):
        if data["GPT4_Eval"] is None:
            print("Skipping entry with None eval results")
            continue

        candidate = {
            "CogVLM": clean_text(data["CogVLM"]),
            "ShareCaptioner": clean_text(data["ShareCaptioner"]),
            "LLaVA_v15": clean_text(data["LLaVA_v15"]),
        }
        candidates.append(candidate)

        ref = {
            "GT_Caption_GPT4O": clean_text(data["GT_Caption_GPT4O"]),
            "GT_Caption_GPT4V": clean_text(data["GT_Caption_GPT4V"]),
            "GT_Caption_Gemini15Pro": clean_text(data["GT_Caption_Gemini15Pro"]),
        }
        refs.append(ref)

        human_score = {
            "CogVLM": json.loads(data["GPT4_Eval"])["CogVLM"],
            "ShareCaptioner": json.loads(data["GPT4_Eval"])["ShareCaptioner"],
            "LLaVA_v15": json.loads(data["GPT4_Eval"])["LLaVA_v15"],
        }
        human_scores.append(human_score)

    print(f"Processed {len(refs)} references and {len(candidates)} candidates")
    return refs, candidates, human_scores


def load_longparse_dataset(
    dataset_name_or_path="XXX",
):
    dataset = load_dataset("json", data_files=dataset_name_or_path, split="train")
    refs = []
    candidates = []
    human_scores = []

    # Process dataset
    for data in tqdm(dataset, desc="Processing dataset"):
        candidates.append(clean_text(data["caption"]))
        refs.append(clean_text(data["scene_graph"]))
        human_scores.append(1.0)

    return refs, candidates, human_scores


def load_sft_prompted_longparse_dataset(
    dataset_name_or_path="XXX",
):
    dataset = load_dataset("json", data_files=dataset_name_or_path, split="train")
    refs = []
    candidates = []
    human_scores = []

    # Process dataset
    for data in tqdm(dataset, desc="Processing dataset"):
        cand = clean_text(data["generated_scene_graph"])
        cand_triples = extract_triples(cand)
        if len(cand_triples) > 1:
            for triple in cand_triples:
                if (
                    "(" in triple[0]
                    or ")" in triple[0]
                    or "(" in triple[1]
                    or ")" in triple[1]
                    or "(" in triple[2]
                    or ")" in triple[2]
                ):
                    cand_triples.remove(triple)
            cand_triples = " , ".join(
                [
                    f"( {triple[0]} , {triple[1]} , {triple[2]} )"
                    for triple in cand_triples
                ]
            )
            candidates.append(cand_triples)
        else:
            candidates.append("")
        refs.append(clean_text(data["scene_graph"]))
        human_scores.append(1.0)

    return refs, candidates, human_scores


def load_caparena_dataset(
    dataset_name_or_path="XXX",
):
    dataset = load_dataset("json", data_files=dataset_name_or_path, split="train")
    refs = []
    candidates = []
    human_scores = []

    # Process dataset
    for data in tqdm(dataset, desc="Processing dataset"):
        if data["winner"] == "skip" or data["winner"] == "bad":
            continue
        candidates.append(
            {
                "caption1": clean_text_caparena(data["caption1"]),
                "caption2": clean_text_caparena(data["caption2"]),
            }
        )
        refs.append({"ref": clean_text_caparena(data["ref"])})
        if data["winner"] == data["source1"]:
            human_scores.append({"caption1": 1.0, "caption2": 0.0})
        elif data["winner"] == data["source2"]:
            human_scores.append({"caption1": 0.0, "caption2": 1.0})
        elif data["winner"] == "equal":
            human_scores.append({"caption1": 0.5, "caption2": 0.5})
        else:
            raise Exception("winner is not in source1 or source2")

    return refs, candidates, human_scores


def load_discourseFOIL_dataset(
    dataset_name_or_path="XXX",
):
    dataset = load_dataset("json", data_files=dataset_name_or_path, split="train")
    refs = []
    candidates = []
    human_scores = []

    # Process dataset
    for data in tqdm(dataset, desc="Processing dataset"):
        if data["fixed"] != "true":
            continue
        candidates.append(
            {
                "hall_caption": clean_text(data["hall_caption"]),
                "human_corrected_hall_caption": clean_text(data["human_corrected_hall_caption"]),
            }
        )
        refs.append({"caption": clean_text(data["caption"])})

        human_scores.append({"hall_caption": 0.0, "human_corrected_hall_caption": 1.0})

    return refs, candidates, human_scores


def load_flickr8k_dataset(
    dataset_name_or_path="XXX",
):
    data = {}
    with open(dataset_name_or_path) as f:
        data.update(json.load(f))
    print('Loaded {} images'.format(len(data)))

    refs = []
    candidates = []
    human_scores = []
    for k, v in list(data.items()):
        for human_judgement in v['human_judgement']:
            if np.isnan(human_judgement['rating']):
                print('NaN')
                continue

            candidate = ' '.join(human_judgement['caption'].split())
            candidates.append(candidate)

            ref = [' '.join(gt.split()) for gt in v['ground_truth']]
            refs.append(ref)
            human_scores.append(human_judgement['rating'])
    print('Loaded {} references and {} candidates'.format(len(refs), len(candidates)))
    assert len(candidates) == len(refs)

    return refs, candidates, human_scores
